# -*- coding: latin-1 -*-

""" Date/Time string parsing module.

    Note about the Y2K problems:

       The parser can only handle years with at least 2 digits. 2
       digit year values get expanded by adding the century using
       DateTime.add_century(), while 3 digit year get converted
       literally. To have 2 digit years also be interpreted literally,
       add leading zeros, e.g. year 99 must be written as 099 or 0099.

    See http://en.wikipedia.org/wiki/Calendar_date for some
    information about various formats used around the world.

    Copyright (c) 1998-2000, Marc-Andre Lemburg; mailto:mal@lemburg.com
    Copyright (c) 2000-2013, eGenix.com Software GmbH; mailto:info@egenix.com
    See the documentation for further information on copyrights,
    or contact the author. All Rights Reserved.

"""
import types,re
import DateTime,ISO,ARPA,Timezone

### Globals

# Enable to produce debugging output
_debug = 0

### Helpers

def _invdict(mapping):
    d =  {}
    for name, code in mapping.iteritems():
        d[code] = name
    return d

### Parsers

# REs for matching date and time parts in a string; These REs
# parse a superset of ARPA, ISO, American and European style dates.
# Timezones are supported via the Timezone submodule.

_year = '(?P<year>-?\d\d+(?!:))'
_fullyear = '(?P<year>-?\d\d\d+(?!:))'
_epoch = '(?P<epoch> *[ABCDE\.]+)'
_year_epoch = '(?:' + _year + _epoch + '?)'
_fullyear_epoch = '(?:' + _fullyear + _epoch + '?)'
_relyear = '(?:\((?P<relyear>[-+]?\d+)\))'

_month = '(?P<month>\d?\d(?!:))'
_fullmonth = '(?P<month>\d\d(?!:))'
_litmonth = ('(?P<litmonth>'
             'jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec|'
             'mär|mae|mrz|mai|okt|dez|'
             'fev|avr|juin|juil|aou|aoû|déc|'
             'ene|abr|ago|dic|'
             'out'
             ')[a-z,\.;]*')
litmonthtable = {
    # English
    'jan':1, 'feb':2, 'mar':3, 'apr':4, 'may':5, 'jun':6,
    'jul':7, 'aug':8, 'sep':9, 'oct':10, 'nov':11, 'dec':12,
    # German
    'mär':3, 'mae':3, 'mrz':3, 'mai':5, 'okt':10, 'dez':12,
    # French
    'fev':2, 'avr':4, 'juin':6, 'juil':7, 'aou':8, 'aoû':8,
    'déc':12,
    # Spanish
    'ene':1, 'abr':4, 'ago':8, 'dic':12,
    # Portuguese
    'out':10,
    }
_relmonth = '(?:\((?P<relmonth>[-+]?\d+)\))'

_day = '(?P<day>\d\d?(?!:))'
_usday = '(?:(?P<day>\d\d?(?!:))(?:st|nd|rd|th|[,.;])?)'
_fullday = '(?P<day>\d\d(?!:))'
_litday = ('(?P<litday>'
           'mon|tue|wed|thu|fri|sat|sun|'
           'die|mit|don|fre|sam|son|'
           'lun|mar|mer|jeu|ven|sam|dim|'
           'mie|jue|vie|sab|dom|'
           'pri|seg|ter|cua|qui'
           ')[a-z]*')
litdaytable = {
    # English
    'mon':0, 'tue':1, 'wed':2, 'thu':3, 'fri':4, 'sat':5, 'sun':6,
    # German
    'die':1, 'mit':2, 'don':3, 'fre':4, 'sam':5, 'son':6,
    # French
    'lun':0, 'mar':1, 'mer':2, 'jeu':3, 'ven':4, 'sam':5, 'dim':6,
    # Spanish
    'mie':2, 'jue':3, 'vie':4, 'sab':5, 'dom':6,
    # Portuguese
    'pri':0, 'seg':1, 'ter':2, 'cua':3, 'qui':4,
    }
_relday = '(?:\((?P<relday>[-+]?\d+)\))'

_hour = '(?P<hour>[012]?\d)'
_minute = '(?P<minute>[0-6]\d)'
_second = '(?P<second>[0-6]\d(?:[.,]\d+)?)'

_days = '(?P<days>\d+(?:[.,]\d+)?)'
_hours = '(?P<hours>\d+(?:[.,]\d+)?)'
_minutes = '(?P<minutes>\d+(?:[.,]\d+)?)'
_seconds = '(?P<seconds>\d+(?:[.,]\d+)?)'

_reldays = '(?:\((?P<reldays>[-+]?\d+(?:[.,]\d+)?)\))'
_relhours = '(?:\((?P<relhours>[-+]?\d+(?:[.,]\d+)?)\))'
_relminutes = '(?:\((?P<relminutes>[-+]?\d+(?:[.,]\d+)?)\))'
_relseconds = '(?:\((?P<relseconds>[-+]?\d+(?:[.,]\d+)?)\))'

_sign = '(?:(?P<sign>[-+]) *)'
_week = 'W(?P<week>\d?\d)'
_zone = Timezone.zone
_ampm = '(?P<ampm>am|pm|a.m.?|p.m.?)'

_time = (_hour + ':' + _minute + '(?::' + _second + ')? *'
         + _ampm + '? *' + _zone + '?')
_isotime = _hour + ':?' + _minute + ':?' + _second + '? *' + _zone + '?'

_weekdate = _year + '-?(?:' + _week + '-?' + _day + '?)?'
_eurodate = _day + '\.' + _month + '\.' + _year_epoch + '?'
_alteuro1date = _day + '/' + _month + '/' + _year_epoch + '?'
_alteuro2date = _day + '-' + _month + '-' + _year_epoch + '?'
_usdate = _month + '/' + _day + '(?:/' + _year_epoch + '|[^/]|$)'
_altusdate = _month + '-' + _day + '-' + _fullyear_epoch
_isodate = _year + '-' + _month + '-?' + _day + '?(?!:)'
_altisodate = _year + _fullmonth + _fullday + '(?!:)'
_usisodate = _fullyear + '/' + _fullmonth + '/' + _fullday
_litdate = ('(?:'+ _litday + '[,.; ]+)?' + 
            _usday + '[- ]+' + 
            _litmonth + '[- ,.a-z]*' +
            _year_epoch + '?')
_altlitdate = ('(?:'+ _litday + '[,.; ]+)? *' + 
               _litmonth + '[ ,.a-z]+' +
               _usday + 
               '(?:[ a-z]+' + _year_epoch + ')?')

_relany = '[*%?a-zA-Z]+'

_relisodate = ('(?:(?:' + _relany + '|' + _year + '|' + _relyear + ')-' +
               '(?:' + _relany + '|' + _month + '|' + _relmonth + ')-' +
               '(?:' + _relany + '|' + _day + '|' + _relday + '))')

_relisotime = ('(?:(?:' + _relany + '|' + _hour + '|' + _relhours + '):' +
               '(?:' + _relany + '|' + _minute + '|' + _relminutes + ')' +
               '(?::(?:' + _relany + '|' + _second + '|' + _relseconds + '))?)')

_isodelta1 = (_sign + '?' +
              _days + ':' + _hours + ':' + _minutes + ':' + _seconds)
_isodelta2 = (_sign + '?' + 
              _hours + ':' + _minutes + ':' + _seconds)
_isodelta3 = (_sign + '?' + 
              _hours + ':' + _minutes)
_litdelta = (_sign + '?' +
             '(?:' + _days + ' *d[a-z]*[,; ]*)?' + 
             '(?:' + _hours + ' *h[a-z]*[,; ]*)?' + 
             '(?:' + _minutes + ' *m[a-z]*[,; ]*)?' +
             '(?:' + _seconds + ' *s[a-z]*[,; ]*)?')
_litdelta2 = (_sign + '?' +
             '(?:' + _days + ' *d[a-z]*[,; ]*)?' + 
              _hours + ':' + _minutes + '(?::' + _seconds + ')?')

_timeRE = re.compile(_time, re.I)
_isotimeRE = re.compile(_isotime, re.I)
_isodateRE = re.compile(_isodate, re.I)
_altisodateRE = re.compile(_altisodate, re.I)
_usisodateRE = re.compile(_usisodate, re.I)
_eurodateRE = re.compile(_eurodate, re.I)
_alteuro1dateRE = re.compile(_alteuro1date, re.I)
_alteuro2dateRE = re.compile(_alteuro2date, re.I)
_usdateRE = re.compile(_usdate, re.I)
_altusdateRE = re.compile(_altusdate, re.I)
_litdateRE = re.compile(_litdate, re.I)
_altlitdateRE = re.compile(_altlitdate, re.I)
_relisodateRE = re.compile(_relisodate, re.I)
_isodelta1RE = re.compile(_isodelta1)
_isodelta2RE = re.compile(_isodelta2)
_isodelta3RE = re.compile(_isodelta3)
_litdeltaRE = re.compile(_litdelta)
_litdelta2RE = re.compile(_litdelta2)
_relisotimeRE = re.compile(_relisotime, re.I)

### Available date parsers

## Most common date order: day month year

# dd.mm.yyyy
DATE_PARSER_DMY1 = 'dmy1'
# dd/mm/yyyy
DATE_PARSER_DMY2 = 'dmy2'
# dd-mm-yyyy
DATE_PARSER_DMY3 = 'dmy3'

# DMY parser class
DATE_DMY_PARSERS = (
    DATE_PARSER_DMY1,
    DATE_PARSER_DMY2,
    DATE_PARSER_DMY3,
    )

## US date order: month day year

# mm/dd/yyyy
DATE_PARSER_MDY1 = 'mdy1'
# mm-dd-yyyy
DATE_PARSER_MDY2 = 'mdy2'

# MDY parser class
DATE_MDY_PARSERS = (
    DATE_PARSER_MDY1,
    DATE_PARSER_MDY2,
    )

## Literal date forms

# Monday, 2nd November 1998
DATE_PARSER_LITERAL1 = 'lit1'
# Monday, November 2nd 1998
DATE_PARSER_LITERAL2 = 'lit2'

# Literal parser class
DATE_LITERAL_PARSERS = (
    DATE_PARSER_LITERAL1,
    DATE_PARSER_LITERAL2,
    )

## ISO date forms

# yyyy-mm-dd
DATE_PARSER_YMD1 = 'ymd1'
# yyyymmdd
DATE_PARSER_YMD2 = 'ymd2'
# yyyy/mm/dd
DATE_PARSER_YMD3 = 'ymd3'

# YMD parser class
DATE_YMD_PARSERS = (
    DATE_PARSER_YMD1,
    DATE_PARSER_YMD2,
    DATE_PARSER_YMD3,
    )

## Unknown formats

DATE_PARSER_UNKNOWN = 'unknown'

### Date parser name mappings

# Mapping name to code
DATE_PARSERS = {
    # New names in mxDateTime 3.2
    'dmy1': DATE_PARSER_DMY1,
    'dmy2': DATE_PARSER_DMY2,
    'dmy3': DATE_PARSER_DMY3,
    'mdy1': DATE_PARSER_MDY1,
    'mdy2': DATE_PARSER_MDY2,
    'lit1': DATE_PARSER_LITERAL1,
    'lit2': DATE_PARSER_LITERAL2,
    'ymd1': DATE_PARSER_YMD1,
    'ymd2': DATE_PARSER_YMD2,
    'ymd3': DATE_PARSER_YMD3,
    'unknown': DATE_PARSER_UNKNOWN,

    # Aliases
    'dmy': DATE_PARSER_DMY1,
    'dmy-dot': DATE_PARSER_DMY1,
    'dmy-hyphen': DATE_PARSER_DMY2,
    'dmy-slash': DATE_PARSER_DMY3,
    'mdy': DATE_PARSER_MDY1,
    'mdy-slash': DATE_PARSER_MDY1,
    'mdy-hyphen': DATE_PARSER_MDY2,
    'iso': DATE_PARSER_YMD1,
    'iso-hyphen': DATE_PARSER_YMD1,
    'iso-compact': DATE_PARSER_YMD2,
    'iso-no-hyphen': DATE_PARSER_YMD2,
    'iso-slash': DATE_PARSER_YMD3,

    # Old names used in mxDateTime <3.1
    'euro': DATE_PARSER_DMY1,
    'alteuro1': DATE_PARSER_DMY2,
    'alteuro2': DATE_PARSER_DMY3,
    'us': DATE_PARSER_MDY1,
    'altus': DATE_PARSER_MDY2,
    'lit': DATE_PARSER_LITERAL1,
    'altlit': DATE_PARSER_LITERAL2,
    'eurlit': DATE_PARSER_LITERAL1, # eurlit was replaced by an extended lit
    'iso': DATE_PARSER_YMD1,
    'altiso': DATE_PARSER_YMD2,
    'usiso': DATE_PARSER_YMD3,
    }

DATE_PARSER_NAME = DATE_PARSERS

DATE_PARSER_RE = {
    # New names in mxDateTime 3.2
    DATE_PARSER_DMY1: _eurodateRE,
    DATE_PARSER_DMY2: _alteuro1dateRE,
    DATE_PARSER_DMY3: _alteuro2dateRE,
    DATE_PARSER_MDY1: _usdateRE,
    DATE_PARSER_MDY2: _altusdateRE,
    DATE_PARSER_LITERAL1: _litdateRE,
    DATE_PARSER_LITERAL2: _altlitdateRE,
    DATE_PARSER_YMD1: _isodateRE,
    DATE_PARSER_YMD2: _altisodateRE,
    DATE_PARSER_YMD3: _usisodateRE,
    DATE_PARSER_UNKNOWN: None,
    }

# Default value for the formats parameter (this also defines the
# order in which the parsers are tried)
DEFAULT_DATE_FORMATS = (
    DATE_PARSER_DMY1,
    DATE_PARSER_YMD3, # Placed here to avoid conflicts with MDY1
    DATE_PARSER_MDY1,
    DATE_PARSER_MDY2,
    DATE_PARSER_YMD1,
    DATE_PARSER_YMD2,
    DATE_PARSER_DMY2,
    DATE_PARSER_DMY3,
    DATE_PARSER_LITERAL1,
    DATE_PARSER_LITERAL2,
    DATE_PARSER_UNKNOWN,
    )

### Available time parsers

# HH:MM or HH:MM:SS or HH:MM:SS.SS
TIME_PARSER_STANDARD = 'standard'

# HHMM or HHMMSS or HHMMSS.SS
TIME_PARSER_ISO = 'iso'

# Unknown format
TIME_PARSER_UNKNOWN = 'unknown'

### Time parser name mapppings

# Map names to codes
TIME_PARSERS = {
    'standard': TIME_PARSER_STANDARD,
    'iso': TIME_PARSER_ISO,
    'unknown': TIME_PARSER_UNKNOWN,
    }

TIME_PARSER_NAME = TIME_PARSERS

# Default value for the time_formats parameter (this also defines the
# order in which the parsers are tried)
DEFAULT_TIME_FORMATS = (
    TIME_PARSER_STANDARD,
    TIME_PARSER_ISO,
    TIME_PARSER_UNKNOWN,
    )

###

def _parse_date(text, date_formats=DEFAULT_DATE_FORMATS, defaultdate=None,

                int=int,float=float,
                add_century=DateTime.add_century,
                now=DateTime.now):

    """ Parses the date part given in text and returns a tuple
        (text,day,month,year,style) with the following
        meanings:

        * text gives the original text without the date part

        * day,month,year give the parsed date

        * style gives information about which parser was successful
          in form of the parser code (see DATE_PARSERS)

        date_formats may be set to a tuple of style codes specifying
        which of the available parsers to use and in which order to
        try them. Default is to try all available formats.

        defaultdate provides the defaults to use in case no date part
        is found. Most other parsers default to the current year
        January 1 if some of these date parts are missing.

        If 'unknown' is not given in formats and the date cannot be
        parsed, a ValueError is raised.

    """
    match = None
    style = 0
    
    # Apply parsers in the order given in formats
    for format_code in date_formats:

        # Find parser
        try:
            parser_re = DATE_PARSER_RE[format_code]
        except KeyError:
            raise TypeError('unknown date format: %r' % format_code)
        
        # Handle DATE_PARSER_UNKNOWN: use defaultdate
        if parser_re is None:
            if defaultdate is None:
                defaultdate = now()
            year = defaultdate.year
            month = defaultdate.month
            day = defaultdate.day
            style = format_code
            break

        # Run parser
        match = parser_re.search(text)
        if match is None:
            continue
        if _debug:
            print 'parser groups: %r' % (match.groups(),)

        # Post-processing
        if format_code in DATE_DMY_PARSERS:

            day,month,year,epoch = match.groups()
            month = int(month)
            # Could have mistaken euro format for us style date
            # which uses month, day order
            if month > 12 or month == 0:
                match = None
                continue
            day = int(day)
            if year:
                if len(year) == 2:
                    # Y2K problem:
                    year = add_century(int(year))
                else:
                    year = int(year)
            else:
                if defaultdate is None:
                    defaultdate = now()
                year = defaultdate.year
            if epoch and 'B' in epoch:
                year = -year + 1
            break

        elif format_code in DATE_YMD_PARSERS:

            if format_code == DATE_PARSER_YMD2:
                # Avoid mistaking ISO time parts ('Thhmmss') for dates
                left, right = match.span()
                if left > 0 and \
                   text[left - 1] == 'T':
                    match = None
                    continue

            year,month,day = match.groups()
            if len(year) == 2:
                # Y2K problem:
                year = add_century(int(year))
            else:
                year = int(year)
            # Default to January 1st
            if not month:
                month = 1
            else:
                month = int(month)
            if not day:
                day = 1
            else:
                day = int(day)
            break

        elif format_code in DATE_MDY_PARSERS:

            month,day,year,epoch = match.groups()
            month = int(month)
            # Could have mistaken us format for euro style date
            # which uses day, month order
            if month > 12 or month == 0:
                match = None
                continue
            # Default to 1 if no day is given
            if day:
                day = int(day)
            else:
                day = 1
            if year:
                if len(year) == 2:
                    # Y2K problem:
                    year = add_century(int(year))
                else:
                    year = int(year)
            else:
                if defaultdate is None:
                    defaultdate = now()
                year = defaultdate.year
            if epoch and 'B' in epoch:
                year = -year + 1
            break

        elif format_code in DATE_LITERAL_PARSERS:

            if format_code == DATE_PARSER_LITERAL1:
                litday,day,litmonth,year,epoch = match.groups()
                month = ''
            elif format_code == DATE_PARSER_LITERAL2:
                litday,litmonth,day,year,epoch = match.groups()
                month = ''

            if litmonth:
                litmonth = litmonth.lower()
                try:
                    month = litmonthtable[litmonth]
                except KeyError:
                    # Unknown month name
                    match = None
                    continue
            elif month:
                month = int(month)
            else:
                month = 1
            # Could have mistaken us format for euro style date
            # which uses day, month order
            if month > 12 or month == 0:
                match = None
                continue

            if day:
                day = int(day)
            else:
                day = 1

            # Default to current year, January 1st
            if not year:
                if defaultdate is None:
                    defaultdate = now()
                year = defaultdate.year
            else:
                if len(year) == 2:
                    # Y2K problem:
                    year = add_century(int(year))
                else:
                    year = int(year)
            if epoch and 'B' in epoch:
                year = -year + 1
            break

    # Check success
    if match is not None:
        # Remove date from text
        left, right = match.span()
        if _debug:
            print 'parsed date:',repr(text[left:right]),\
                  'giving:',year,month,day
        text = text[:left] + text[right:]
        style = format_code
        
    elif not style:
        # Not recognized: raise an error
        raise ValueError('unknown date format: "%s"' % text)

    #print '_parse_date:',text,day,month,year,style
    return text,day,month,year,style

def _parse_time(text, time_formats=DEFAULT_TIME_FORMATS,

                int=int,float=float):

    """ Parses a time part given in text and returns a tuple
        (text,hour,minute,second,offset,style) with the following
        meanings:

        * text gives the original text without the time part

        * hour,minute,second give the parsed time

        * offset gives the time zone UTC offset

        * style gives information about which parser was successful:
          in form of the parser code (see TIME_PARSERS)

        time_formats may be set to a tuple of style codes specifying
        which of the available parsers to use and in which order to
        try them. Default is to try all available formats.

        If 'unknown' is not given in formats and the time cannot be
        parsed, a ValueError is raised.

    """
    match = None
    style = 0

    # Apply parsers in the order given in formats
    for format_code in time_formats:

        if format_code == TIME_PARSER_STANDARD:

            # Standard format
            match = _timeRE.search(text)
            if match is None:
                continue
            hour,minute,second,ampm,zone = match.groups()
            style = format_code
            break

        elif format_code == TIME_PARSER_ISO:

            # ISO format
            match =  _isotimeRE.search(text)
            if match is None:
                continue
            hour,minute,second,zone = match.groups()
            ampm = None
            style = format_code
            break

        elif format_code == TIME_PARSER_UNKNOWN:

            # Default handling
            hour,minute,second,offset = 0,0,0.0,0
            style = format_code
            break

        else:
            raise TypeError('unknown time format: %r' % format_code)

    # Check success
    if not style:
        # If no default handling should be applied, raise an error
        raise ValueError, 'unknown time format: "%s"' % text

    # Post-processing
    if match is not None:
        if zone:
            # Convert to UTC offset
            offset = Timezone.utc_offset(zone)
        else:
            offset = 0
        hour = int(hour)
        if ampm:
            if ampm[0] in ('p', 'P'):
                # 12pm = midday
                if hour < 12:
                    hour = hour + 12
            else:
                # 12am = midnight 
                if hour >= 12:
                    hour = hour - 12
        if minute:
            minute = int(minute)
        else:
            minute = 0
        if not second:
            second = 0.0
        else:
            if ',' in second:
                second = second.replace(',', '.')
            second = float(second)

        # Remove time from text
        left,right = match.span()
        if _debug: 
            print 'parsed time:',repr(text[left:right]),\
                  'giving:',hour,minute,second,offset
        text = text[:left] + text[right:]

    #print '_parse_time:',text,hour,minute,second,offset,style
    return text,hour,minute,second,offset,style

###

def DateTimeFromString(text, formats=DEFAULT_DATE_FORMATS, defaultdate=None,
                       time_formats=DEFAULT_TIME_FORMATS,

                       DateTime=DateTime):

    """ DateTimeFromString(text, [formats, defaultdate])
    
        Returns a DateTime instance reflecting the date and time given
        in text. In case a timezone is given, the returned instance
        will point to the corresponding UTC time value. Otherwise, the
        value is set as given in the string.

        formats may be set to a tuple of strings specifying which of
        the following parsers to use and in which order to try
        them. Default is to try all of them in the order given below:

          'dmy1' - the European date parser (dd.mm.yyyy)
          'ymd3' - US style ISO date parser (yyyy/mm/dd)
          'mdy1' - the US date parser (mm/dd/yyyy)
          'mdy2' - the alternative US date parser (mm-dd-yyyy)
          'ymd1' - the ISO date parser (yyyy-mm-dd)
          'ymd2' - the alternative ISO date parser (yyyymmdd)
          'dmy2' - alternative European date parser 1 (dd/mm/yyyy)
          'dmy3' - alternative European date parser 2 (dd-mm-yyyy)
          'lit1' - the literal date parser (day, month order)
          'lit2' - the alternative literal date parser (month, day order)
          'unknown' - no date part was found, defaultdate was used

        defaultdate provides the defaults to use in case no date part
        is found. Most of the parsers default to the current year
        January 1 if some of these date parts are missing.

        If 'unknown' is not given in formats and the date cannot
        be parsed, a ValueError is raised.

        time_formats may be set to a tuple of strings specifying which
        of the following parsers to use and in which order to try
        them. Default is to try all of them in the order given below:

          'standard' - standard time format HH:MM:SS
          'iso'      - ISO time format (superset of 'standard')
          'unknown'  - default to 00:00:00 in case the time format
                       cannot be parsed

        Defaults to 00:00:00.00 for time parts that are not included
        in the textual representation.

        If 'unknown' is not given in time_formats and the time cannot
        be parsed, a ValueError is raised.

    """
    origtext = text.strip()

    # Convert formats list to a list of parser codes
    if formats is not DEFAULT_DATE_FORMATS:
        formats = tuple([DATE_PARSERS[format]
                         for format in formats])
        if DATE_PARSER_UNKNOWN not in formats:
            formats_with_unknown = formats + (DATE_PARSER_UNKNOWN,)
        else:
            formats_with_unknown = formats
    else:
        formats_with_unknown = formats

    # Convert time_formats list to a list of parser codes
    if time_formats is not DEFAULT_TIME_FORMATS:
        time_formats = tuple([TIME_PARSERS[format]
                              for format in time_formats])
        if TIME_PARSER_UNKNOWN not in time_formats:
            time_formats_with_unknown = time_formats + (TIME_PARSER_UNKNOWN,)
        else:
            time_formats_with_unknown = time_formats
    else:
        time_formats_with_unknown = time_formats

    # First try: parse date, then time
    if _debug:
        print 'first try on %r using these formats: %r' % (
            origtext, formats_with_unknown)
    text,day,month,year,datestyle = _parse_date(
        origtext,
        formats_with_unknown,
        defaultdate)
    text,hour,minute,second,offset,timestyle = _parse_time(
        text,
        time_formats_with_unknown)
    text = text.strip()
    if _debug:
        print ('result of date/time parsing on %r: '
               'datestyle=%s, timestyle=%s, text=%r' % (
                   origtext,
                   DATE_PARSER_NAME[datestyle],
                   TIME_PARSER_NAME[timestyle],
                   text))
    try1 = (text,
            day,month,year,datestyle,
            hour,minute,second,offset,timestyle)
    try1_text_left = len(text)

    # If we have text left, run a second try, if the date style could
    # not be determined or we have found a literal format
    if (try1_text_left > 0 and
        (datestyle == DATE_PARSER_UNKNOWN or
         datestyle in DATE_LITERAL_PARSERS)):

        # Second try: parse time, then date
        if _debug:
            print 'second try on %r using these formats: %r' % (
                origtext, formats_with_unknown)
        text,hour,minute,second,offset,timestyle = _parse_time(
            origtext,
            time_formats_with_unknown)
        text,day,month,year,datestyle = _parse_date(
            text,
            formats_with_unknown,
            defaultdate)
        text = text.strip()
        if _debug:
            print ('result of time/date parsing on %r: '
                   'datestyle=%s, timestyle=%s, text=%r' % (
                       origtext,
                       DATE_PARSER_NAME[datestyle],
                       TIME_PARSER_NAME[timestyle],
                       text))

        # Decide based on the length of the remaining text and the
        # parsed styles
        try2_text_left = len(text)
        if try1_text_left < try2_text_left:
            try1_wins = 1
        elif try2_text_left > try1_text_left:
            try1_wins = 0
        else:
            # Equal length: use the try with the better parsing
            # characteristics, emphasizing the date part
            if datestyle == DATE_PARSER_UNKNOWN:
                try1_wins = 1
            elif (timestyle == TIME_PARSER_UNKNOWN and
                  try1[4] != DATE_PARSER_UNKNOWN):
                try1_wins = 1
            else:
                try1_wins = 0
        if try1_wins:
            # Try 1 wins
            (text,
             day,month,year,datestyle,
             hour,minute,second,offset,timestyle) = try1
            if _debug:
                print 'first try wins'
        else:
            # Try 2 wins
            if _debug:
                print 'second try wins'

    else:
        if _debug:
            print 'first try succeeds without second try'

    # Final check for unknowns
    if ((datestyle == DATE_PARSER_UNKNOWN and
         formats is not formats_with_unknown) or
        (timestyle == TIME_PARSER_UNKNOWN and
         time_formats is not time_formats_with_unknown)):
        raise ValueError(
            'Failed to parse "%s": found "%s" date, "%s" time' %
            (origtext,
             DATE_PARSER_NAME[datestyle],
             TIME_PARSER_NAME[timestyle]))

    # Convert to DateTime instance
    try:
        return DateTime.DateTime(year,month,day,hour,minute,second) - offset
    except DateTime.RangeError, why:
        raise DateTime.RangeError,\
              'Failed to parse "%s": %s' % (origtext, why)

def DateFromString(text, formats=DEFAULT_DATE_FORMATS, defaultdate=None,

                   DateTime=DateTime):

    """ DateFromString(text, [formats, defaultdate])
    
        Returns a DateTime instance reflecting the date given in
        text. A possibly included time part is ignored.

        formats and defaultdate work just like for
        DateTimeFromString().

    """
    text = text.strip()

    # Convert formats list to a list of parser codes
    if formats is not DEFAULT_DATE_FORMATS:
        formats = tuple([DATE_PARSERS[format]
                         for format in formats])

    # Parse date
    _text,day,month,year,datestyle = _parse_date(text, formats, defaultdate)

    if (datestyle == DATE_PARSER_UNKNOWN and
        DATE_PARSER_UNKNOWN not in formats):
        raise ValueError(
            'Failed to parse "%s": found "%s" date' %
            (text, datestyle))

    try:
        return DateTime.DateTime(year,month,day)
    except DateTime.RangeError, why:
        raise DateTime.RangeError,\
              'Failed to parse "%s": %s' % (text, why)

def validateDateTimeString(text, formats=DEFAULT_DATE_FORMATS,
                           time_formats=DEFAULT_TIME_FORMATS):

    """ validateDateTimeString(text, [formats, defaultdate])

        Validates the given text and returns 1/0 depending on whether
        text includes parseable date and time values or not.

        formats works just like for DateTimeFromString() and defines
        the order of date/time parsers to apply. It defaults to the
        same list of parsers as for DateTimeFromString().

        XXX Undocumented !
    
    """
    formats = list(formats)
    if DATE_PARSER_UNKNOWN in formats:
        formats.remove(DATE_PARSER_UNKNOWN)
    time_formats = list(time_formats)
    if TIME_PARSER_UNKNOWN in time_formats:
        time_formats.remove(TIME_PARSER_UNKNOWN)
    try:
        DateTimeFromString(text, formats=formats, time_formats=time_formats)
    except (DateTime.RangeError, ValueError), why:
        return 0
    return 1

def validateDateString(text, formats=DEFAULT_DATE_FORMATS):

    """ validateDateString(text, [formats, defaultdate])

        Validates the given text and returns 1/0 depending on whether
        text includes a parseable date value or not.

        formats works just like for DateTimeFromString() and defines
        the order of date/time parsers to apply. It defaults to the
        same list of parsers as for DateTimeFromString().
    
        XXX Undocumented !
    
    """
    formats = list(formats)
    if DATE_PARSER_UNKNOWN in formats:
        formats.remove(DATE_PARSER_UNKNOWN)
    try:
        DateFromString(text, formats=formats)
    except (DateTime.RangeError, ValueError), why:
        return 0
    return 1

def TimeFromString(text, formats=DEFAULT_TIME_FORMATS,

                   DateTime=DateTime):

    """ TimeFromString(text, [formats])
    
        Returns a DateTimeDelta instance reflecting the time given in
        text. A possibly included date part is ignored.

        formats may be set to a tuple of strings specifying which of
        the following parsers to use and in which order to try
        them. Default is to try all of them in the order given below:

          'standard' - standard time format HH:MM:SS
          'iso'      - ISO time format (superset of 'standard')
          'unknown'  - default to 00:00:00 in case the time format
                       cannot be parsed

        Defaults to 00:00:00.00 for parts that are not included in the
        textual representation.
        
    """
    text = text.strip()

    # Convert time_formats list to a list of parser codes
    if formats is not DEFAULT_TIME_FORMATS:
        formats = tuple([TIME_PARSERS[format]
                         for format in formats])

    # Parse time
    _text,hour,minute,second,offset,timestyle = _parse_time(
        text,
        formats)

    if (timestyle == TIME_PARSER_UNKNOWN and
        TIME_PARSER_UNKNOWN not in formats):
        raise ValueError(
            'Failed to parse "%s": found "%s" time' %
            (text, timestyle))

    try:
        dtd = DateTime.DateTimeDelta(0.0, hour, minute, second)
    except DateTime.RangeError, why:
        raise DateTime.RangeError,\
              'Failed to parse "%s": %s' % (text, why)
    else:
        # XXX What to do with offset ?
        return dtd

#
# XXX Still missing: validateTimeString(), validateDateTimeDeltaString()
#                    and validateTimeDeltaString()
#

def DateTimeDeltaFromString(text,

                            float=float,DateTime=DateTime):

    """ DateTimeDeltaFromString(text)
    
        Returns a DateTimeDelta instance reflecting the delta given in
        text. Defaults to 0:00:00:00.00 for parts that are not
        included in the textual representation or cannot be parsed.

    """
    match = _isodelta1RE.search(text)
    if match is not None:
        sign, days, hours, minutes, seconds = match.groups()
    else:
        match = _litdelta2RE.search(text)
        if match is not None:
            sign, days, hours, minutes, seconds = match.groups()
        else:
            match = _isodelta2RE.search(text)
            if match is not None:
                sign, hours, minutes, seconds = match.groups()
                days = None
            else:
                match = _isodelta3RE.search(text)
                if match is not None:
                    sign, hours, minutes = match.groups()
                    days = None
                    seconds = None
                else:
                    match = _litdeltaRE.search(text)
                    if match is not None:
                        sign, days, hours, minutes, seconds = match.groups()

                    else:
                        # Not matched:
                        return DateTime.DateTimeDelta(0.0)

    # Conversions
    if days:
        days = float(days)
    else:
        days = 0.0
    if hours:
        hours = float(hours)
    else:
        hours = 0.0
    if minutes:
        minutes = float(minutes)
    else:
        minutes = 0.0
    if seconds:
        seconds = float(seconds)
    else:
        seconds = 0.0
    if sign != '-':
        sign = 1
    else:
        sign = -1

    try:
        dtd = DateTime.DateTimeDelta(days,hours,minutes,seconds)
    except DateTime.RangeError, why:
        raise DateTime.RangeError,\
              'Failed to parse "%s": %s' % (text, why)
    else:
        if sign < 0:
            return -dtd
        else:
            return dtd

# Aliases
TimeDeltaFromString = DateTimeDeltaFromString

###

def _parse_reldate(text,

                   int=int,float=float):

    match = _relisodateRE.search(text)
    if match is not None:
        groups = match.groups()
        if _debug: print groups
        year,years,month,months,day,days = groups
        if year:
            year = int(year)
        if years:
            years = float(years)
        else:
            years = 0
        if month:
            month = int(month)
        if months:
            months = float(months)
        else:
            months = 0
        if day:
            day = int(day)
        if days:
            days = float(days)
        else:
            days = 0
        return year,years,month,months,day,days
    else:
        return None,0,None,0,None,0

def _parse_reltime(text,

                   int=int,float=float):

    match = _relisotimeRE.search(text)
    if match is not None:
        groups = match.groups()
        if _debug: print groups
        hour,hours,minute,minutes,second,seconds = groups
        if hour:
            hour = int(hour)
        if hours:
            hours = float(hours)
        else:
            hours = 0
        if minute:
            minute = int(minute)
        if minutes:
            minutes = float(minutes)
        else:
            minutes = 0
        if second:
            second = int(second)
        if seconds:
            seconds = float(seconds)
        else:
            seconds = 0
        return hour,hours,minute,minutes,second,seconds
    else:
        return None,0,None,0,None,0

def RelativeDateTimeFromString(text,

                               RelativeDateTime=DateTime.RelativeDateTime):

    """ RelativeDateTimeFromString(text)
    
        Returns a RelativeDateTime instance reflecting the relative
        date and time given in text.

        Defaults to wildcards for parts or values which are not
        included in the textual representation or cannot be parsed.

        The format used in text must adhere to the following syntax:

        		[YYYY-MM-DD] [HH:MM[:SS]]

        with the usual meanings. Values which should not be altered
        may be replaced with '*', '%', '?' or any combination of
        letters, e.g. 'YYYY'. Relative settings must be enclosed in
        parenthesis if given and should include a sign, e.g. '(+0001)'
        for the year part. All other settings are interpreted as
        absolute values.

        Date and time parts are both optional as a whole. Seconds in
        the time part are optional too. Everything else (including the
        hyphens and colons) is mandatory.

    """
    text = text.strip()

    year,years,month,months,day,days = _parse_reldate(text)
    hour,hours,minute,minutes,second,seconds = _parse_reltime(text)
    return RelativeDateTime(year=year,years=years,
                            month=month,months=months,
                            day=day,days=days,
                            hour=hour,hours=hours,
                            minute=minute,minutes=minutes,
                            second=second,seconds=seconds)

def RelativeDateFromString(text,

                           RelativeDateTime=DateTime.RelativeDateTime):

    """ RelativeDateFromString(text)
    
        Same as RelativeDateTimeFromString(text) except that only the
        date part of text is taken into account.

    """
    text = text.strip()

    year,years,month,months,day,days = _parse_reldate(text)
    return RelativeDateTime(year=year,years=years,
                            month=month,months=months,
                            day=day,days=days)

def RelativeTimeFromString(text,

                           RelativeDateTime=DateTime.RelativeDateTime):

    """ RelativeTimeFromString(text)
    
        Same as RelativeDateTimeFromString(text) except that only the
        time part of text is taken into account.

    """
    text = text.strip()

    hour,hours,minute,minutes,second,seconds = _parse_reltime(text)
    return RelativeDateTime(hour=hour,hours=hours,
                            minute=minute,minutes=minutes,
                            second=second,seconds=seconds)

### Tests

def _test():

    import sys

    now = DateTime.now()

    class T:
        def __init__(self, text, reference, **parser_kws):
            self.text = text
            self.reference = reference
            self.parser_kws = parser_kws

    print 'Testing DateTime Parser...'

    l = [

        # Literal formats
        T('Sun Nov  6 08:49:37 1994', '1994-11-06 08:49:37.00'),
        T('sun nov  6 08:49:37 1994', '1994-11-06 08:49:37.00'),
        T('sUN NOV  6 08:49:37 1994', '1994-11-06 08:49:37.00'),
        T('Sunday, 06-Nov-94 08:49:37 GMT', '1994-11-06 08:49:37.00'),
        T('Sun, 06 Nov 1994 08:49:37 GMT', '1994-11-06 08:49:37.00'),
        T('06-Nov-94 08:49:37', '1994-11-06 08:49:37.00'),
        T('06-Nov-94', '1994-11-06 00:00:00.00'),
        T('06-NOV-94', '1994-11-06 00:00:00.00'),
        T('November 19 08:49:37', '%s-11-19 08:49:37.00' % now.year),
        T('Nov. 9', '%s-11-09 00:00:00.00' % now.year),
        T('Sonntag, der 6. November 1994, 08:49:37 GMT', '1994-11-06 08:49:37.00'),
        T('6. November 2001, 08:49:37', '2001-11-06 08:49:37.00'),
        T('sep 6', '%s-09-06 00:00:00.00' % now.year),
        T('sep 6 2000', '2000-09-06 00:00:00.00'),
        T('September 29', '%s-09-29 00:00:00.00' % now.year),
        T('Sep. 29', '%s-09-29 00:00:00.00' % now.year),
        T('6 sep', '%s-09-06 00:00:00.00' % now.year),
        T('29 September', '%s-09-29 00:00:00.00' % now.year),
        T('29 Sep.', '%s-09-29 00:00:00.00' % now.year),
        T('sep 6 2001', '2001-09-06 00:00:00.00'),
        T('Sep 6, 2001', '2001-09-06 00:00:00.00'),
        T('September 6, 2001', '2001-09-06 00:00:00.00'),
        T('sep 6 01', '2001-09-06 00:00:00.00'),
        T('Sep 6, 01', '2001-09-06 00:00:00.00'),
        T('September 6, 01', '2001-09-06 00:00:00.00'),
        T('Thursday, Sept 6 01', '2001-09-06 00:00:00.00'),
        T('Thursday, 6 Sept 01', '2001-09-06 00:00:00.00'),
        T('Thursday, 2001-09-06', '2001-09-06 00:00:00.00'),
        T('Thursday, 6-09-2001', '2001-09-06 00:00:00.00', formats=['dmy3']),
        T('Thursday, 6-09-2001', '2001-06-09 00:00:00.00'),
        T('Thursday, 6/9/2001', '2001-06-09 00:00:00.00'),
        T('30 Apr 2006 20:19:00', '2006-04-30 20:19:00.00'),
        
        # ISO formats
        T('1994-11-06 08:49:37', '1994-11-06 08:49:37.00'),
        T('010203', '2001-02-03 00:00:00.00'),
        T('2001-02-03 00:00:00.00', '2001-02-03 00:00:00.00'),
        T('2001-02 00:00:00.00', '2001-02-01 00:00:00.00'),
        T('2001-02-03', '2001-02-03 00:00:00.00'),
        T('2001-02', '2001-02-01 00:00:00.00'),
        T('20000824/2300', '2000-08-24 23:00:00.00'),
        T('20000824/0102', '2000-08-24 01:02:00.00'),
        T('20000824', '2000-08-24 00:00:00.00'),
        T('20000824/020301', '2000-08-24 02:03:01.00'),
        T('20000824 020301', '2000-08-24 02:03:01.00'),
        T('-20000824 020301', '-2000-08-24 02:03:01.00'),
        T('20000824T020301', '2000-08-24 02:03:01.00'),
        T('20000824 020301', '2000-08-24 02:03:01.00'),
        T('2000-08-24 02:03:01.00', '2000-08-24 02:03:01.00'),
        T('T020311', '%s 02:03:11.00' % now.date),
        T('2003-12-9', '2003-12-09 00:00:00.00'),
        T('03-12-9', '2003-12-09 00:00:00.00'),
        T('003-12-9', '0003-12-09 00:00:00.00'),
        T('0003-12-9', '0003-12-09 00:00:00.00'),
        T('2003-1-9', '2003-01-09 00:00:00.00'),
        T('03-1-9', '2003-01-09 00:00:00.00'),
        T('003-1-9', '0003-01-09 00:00:00.00'),
        T('0003-1-9', '0003-01-09 00:00:00.00'),
        T('2008-05-06 19:30:09.57', '2008-05-06 19:30:09.57'),
        T('2008-05-06 19:30:09.58', '2008-05-06 19:30:09.58'),
        T('2008-05-06 19:30:09.59', '2008-05-06 19:30:09.59'),
        T('2008-05-06 19:30:09.60', '2008-05-06 19:30:09.60'),
        T('2008-05-06 19:30:09.61', '2008-05-06 19:30:09.61'),
        T('2008-05-06 19:30:09.62', '2008-05-06 19:30:09.62'),
        T('2008-05-06 19:30:09.570', '2008-05-06 19:30:09.57'),
        T('2008-05-06 19:30:09.580', '2008-05-06 19:30:09.58'),
        T('2008-05-06 19:30:09.590', '2008-05-06 19:30:09.59'),
        T('2008-05-06 19:30:09.600', '2008-05-06 19:30:09.60'),
        T('2008-05-06 19:30:09.610', '2008-05-06 19:30:09.61'),
        T('2008-05-06 19:30:09.620', '2008-05-06 19:30:09.62'),
        T('2008-05-06 19:30:09.571', '2008-05-06 19:30:09.57'),
        T('2008-05-06 19:30:09.581', '2008-05-06 19:30:09.58'),
        T('2008-05-06 19:30:09.591', '2008-05-06 19:30:09.59'),
        T('2008-05-06 19:30:09.601', '2008-05-06 19:30:09.60'),
        T('2008-05-06 19:30:09.611', '2008-05-06 19:30:09.61'),
        T('2008-05-06 19:30:09.621', '2008-05-06 19:30:09.62'),
        T('2008-05-06 19:30:09.575', '2008-05-06 19:30:09.58'),
        T('2008-05-06 19:30:09.585', '2008-05-06 19:30:09.59'),
        T('2008-05-06 19:30:09.595', '2008-05-06 19:30:09.60'),
        T('2008-05-06 19:30:09.605', '2008-05-06 19:30:09.61'),
        T('2008-05-06 19:30:09.615', '2008-05-06 19:30:09.62'),
        T('2008-05-06 19:30:09.625', '2008-05-06 19:30:09.63'),
        T('2008-05-06 19:30:09.579', '2008-05-06 19:30:09.58'),
        T('2008-05-06 19:30:09.589', '2008-05-06 19:30:09.59'),
        T('2008-05-06 19:30:09.599', '2008-05-06 19:30:09.60'),
        T('2008-05-06 19:30:09.609', '2008-05-06 19:30:09.61'),
        T('2008-05-06 19:30:09.619', '2008-05-06 19:30:09.62'),
        T('2008-05-06 19:30:09.629', '2008-05-06 19:30:09.63'),

        # US formats
        T('06/11/94 08:49:37', '1994-06-11 08:49:37.00'),
        T('11/06/94 08:49:37', '1994-11-06 08:49:37.00'),
        T('9/23/2001', '2001-09-23 00:00:00.00'),
        T('9-23-2001', '2001-09-23 00:00:00.00'),
        T('9/6', '%s-09-06 00:00:00.00' % now.year),
        T('09/6', '%s-09-06 00:00:00.00' % now.year),
        T('9/06', '%s-09-06 00:00:00.00' % now.year),
        T('09/06', '%s-09-06 00:00:00.00' % now.year),
        T('9/6/2001', '2001-09-06 00:00:00.00'),
        T('09/6/2001', '2001-09-06 00:00:00.00'),
        T('9/06/2001', '2001-09-06 00:00:00.00'),
        T('09/06/2001', '2001-09-06 00:00:00.00'),
        T('9-6-2001', '2001-09-06 00:00:00.00'),
        T('09-6-2001', '2001-09-06 00:00:00.00'),
        T('9-06-2001', '2001-09-06 00:00:00.00'),
        T('09-06-2001', '2001-09-06 00:00:00.00'),
        T('2002/05/28 13:10:56.1147 GMT+2', '2002-05-28 13:10:56.11'),
        T('1970/01/01', '1970-01-01 00:00:00.00'),
        T('20021025 12:00 PM', '2002-10-25 12:00:00.00'),
        T('20021025 12:30 PM', '2002-10-25 12:30:00.00'),
        T('20021025 12:00 AM', '2002-10-25 00:00:00.00'),
        T('20021025 12:30 AM', '2002-10-25 00:30:00.00'),
        T('20021025 1:00 PM', '2002-10-25 13:00:00.00'),
        T('20021025 2:00 AM', '2002-10-25 02:00:00.00'),
        T('Thursday, February 06, 2003 12:40 PM', '2003-02-06 12:40:00.00'),
        T('Mon, 18 Sep 2006 23:03:00', '2006-09-18 23:03:00.00'),
        T('6/12/08 5:08pm', '2008-06-12 17:08:00.00'),
        T('6/12/08 5:08p.m.', '2008-06-12 17:08:00.00'),
        T('6/12/08 5:08 pm', '2008-06-12 17:08:00.00'),
        T('6/12/08 5:08 p.m.', '2008-06-12 17:08:00.00'),
        T('January 12th 2008 at 5:08pm', '2008-01-12 17:08:00.00'),
        T('January 12th 2008 at 5:08 pm', '2008-01-12 17:08:00.00'),

        # European formats
        T('6.11.2001, 08:49:37', '2001-11-06 08:49:37.00'),
        T('06.11.2001, 08:49:37', '2001-11-06 08:49:37.00'),
        T('06.11. 08:49:37', '%s-11-06 08:49:37.00' % now.year),
        T('10/12/2004', '2004-12-10 00:00:00.00', formats=['dmy2']),
        T('13/02/2009', '2009-02-13 00:00:00.00', formats=['dmy2']),
        T('21/12/2002', '2002-12-21 00:00:00.00'),
        T('21/08/2002', '2002-08-21 00:00:00.00'),
        T('21-08-2002', '2002-08-21 00:00:00.00', formats=['dmy3']),
        T('13/01/03', '2003-01-13 00:00:00.00'),
        #T('13/1/03', '2003-01-13 00:00:00.00', formats=['dmy2']),
        #T('13/1/3', '2003-01-13 00:00:00.00', formats=['dmy2']),
        #T('13/01/3', '2003-01-13 00:00:00.00'),

        # Time only formats
        T('01:03', '%s 01:03:00.00' % now.date),
        T('01:03:11', '%s 01:03:11.00' % now.date),
        T('01:03:11.50', '%s 01:03:11.50' % now.date),
        T('01:03:11.50 AM', '%s 01:03:11.50' % now.date),
        T('01:03:11.50 PM', '%s 13:03:11.50' % now.date),
        T('01:03:11.50 a.m.', '%s 01:03:11.50' % now.date),
        T('01:03:11.50 p.m.', '%s 13:03:11.50' % now.date),

        # Invalid formats
        T('6..2001, 08:49:37', '%s 08:49:37.00' % now.date),
        T('9//2001', 'ignore'),
        T('06--94 08:49:37', 'ignore'),
        T('20000824020301', 'ignore'),
        T('20-03 00:00:00.00', 'ignore'),
        T('9/2001', 'ignore'),
        T('9-6', 'ignore'),
        T('09-6', 'ignore'),
        T('9-06', 'ignore'),
        T('09-06', 'ignore'),
        T('20000824/23', 'ignore'),
        T('November 1994 08:49:37', 'ignore'),
        T('Nov. 94', 'ignore'),
        T('Mon, 18 Sep 2006 23:03:00 +1234567890', 'ignore'),

        ]

    # Add Unicode versions
    try:
        unicode
    except NameError:
        pass
    else:
        k = []
        for test_case in l:
            k.append(T(unicode(test_case.text),
                       test_case.reference,
                       **test_case.parser_kws))
        l.extend(k)

    t = DateTime.now()

    for test_case in l:
        text = test_case.text
        reference = test_case.reference
        parser_kws = test_case.parser_kws
        try:
            value = DateTimeFromString(text, **parser_kws)
        except Exception, reason:
            if reference is None:
                continue
            else:
                value = str(reason)
        valid_datetime = validateDateTimeString(text)
        valid_date = validateDateString(text)
        if str(value) != reference and \
           not reference == 'ignore':
            print 'Failed to parse "%s"' % text
            print '  expected: %s' % (reference or '<exception>')
            print '  parsed:   %s' % value
            print '  %r' % (_parse_date(
                text,
                date_formats=tuple(parser_kws.get('formats',
                                                  DEFAULT_DATE_FORMATS))
                             + (DATE_PARSER_UNKNOWN,)),)
        elif _debug:
            print 'Parsed "%s" successfully' % text
        if _debug:
            if not valid_datetime:
                print '  "%s" failed date/time validation' % text
            if not valid_date:
                print '  "%s" failed date validation' % text

    et = DateTime.now()
    print 'done. (after %f seconds)' % ((et-t).seconds)

    ###

    print 'Testing DateTimeDelta Parser...'

    l = [

        # Literal formats
        ('Sun Nov  6 08:49:37 1994', '08:49:37.00'),
        ('1 day, 8 hours, 49 minutes, 37 seconds', '1:08:49:37.00'),
        ('10 days, 8 hours, 49 minutes, 37 seconds', '10:08:49:37.00'),
        ('8 hours, 49 minutes, 37 seconds', '08:49:37.00'),
        ('49 minutes, 37 seconds', '00:49:37.00'),
        ('37 seconds', '00:00:37.00'),
        ('37.5 seconds', '00:00:37.50'),
        ('8 hours later', '08:00:00.00'),
        ('2 days', '2:00:00:00.00'),
        ('2 days 23h', '2:23:00:00.00'),
        ('2 days 23:57', '2:23:57:00.00'),
        ('2 days 23:57:13', '2:23:57:13.00'),
        ('', '00:00:00.00'),
        
        # ISO formats
        ('1994-11-06 08:49:37', '08:49:37.00'),
        ('10:08:49:37', '10:08:49:37.00'),
        ('08:49:37', '08:49:37.00'),
        ('08:49', '08:49:00.00'),
        ('-10:08:49:37', '-10:08:49:37.00'),
        ('-08:49:37', '-08:49:37.00'),
        ('-08:49', '-08:49:00.00'),
        ('- 10:08:49:37', '-10:08:49:37.00'),
        ('- 08:49:37', '-08:49:37.00'),
        ('- 08:49', '-08:49:00.00'),
        ('10:08:49:37.5', '10:08:49:37.50'),
        ('08:49:37.5', '08:49:37.50'),
        ('10:8:49:37', '10:08:49:37.00'),
        ('8:9:37', '08:09:37.00'),
        ('8:9', '08:09:00.00'),
        ('8', '00:00:00.00'),

        # Invalid formats
        #('', None),
        #('8', None),

        ]

    t = DateTime.now()

    for text, reference in l:
        try:
            value = DateTimeDeltaFromString(text)
        except:
            if reference is None:
                continue
            else:
                value = str(sys.exc_info()[1])
        if str(value) != reference and \
           not reference == 'ignore':
            print 'Failed to parse "%s"' % text
            print '  expected: %s' % (reference or '<exception>')
            print '  parsed:   %s' % value
        elif _debug:
            print 'Parsed "%s" successfully' % text

    et = DateTime.now()
    print 'done. (after %f seconds)' % ((et-t).seconds)

    ###

    print 'Testing Time Parser...'

    l = [

        # Standard formats
        ('08:49:37 AM', '08:49:37.00'),
        ('08:49:37 PM', '20:49:37.00'),
        ('12:00:00 AM', '00:00:00.00'),
        ('12:00:00 PM', '12:00:00.00'),
        ('8:09:37', '08:09:37.00'),
        ('8:09', '08:09:00.00'),
        
        # ISO formats
        ('08:49:37', '08:49:37.00'),
        ('08:49', '08:49:00.00'),
        ('08:49:37.5', '08:49:37.50'),
        ('08:49:37,5', '08:49:37.50'),
        ('08:09', '08:09:00.00'),

        # Invalid formats
        ('', None),
        ('8:9:37', 'XXX Should give an exception'),
        ('08:9:37', 'XXX Should give an exception'),
        ('8:9', None),
        ('8', None),

        ]

    t = DateTime.now()

    for text, reference in l:
        try:
            value = TimeFromString(text, formats=('standard', 'iso'))
        except:
            if reference is None:
                continue
            else:
                value = str(sys.exc_info()[1])
        if str(value) != reference and \
           not reference == 'ignore':
            print 'Failed to parse "%s"' % text
            print '  expected: %s' % (reference or '<exception>')
            print '  parsed:   %s' % value
        elif _debug:
            print 'Parsed "%s" successfully' % text

    et = DateTime.now()
    print 'done. (after %f seconds)' % ((et-t).seconds)

if __name__ == '__main__':
    _test()



from mx.DateTime import *

c = 35000.0
error = 0
for i in range(1000000):
    t = DateTimeFromCOMDate(c)
    tc = t.COMDate()
    if not tc == c:
        print 'failed for: %s (%.16f != %.16f)' % (t,c,tc)
        error = 1
    # Let's give IEEE floats a hard time...
    if i % 10 == 0:
        c = c - 0.009
    else:
        c = c + 0.0099

if not error:
    print 'All went well !'



#!/usr/bin/env python

""" Distutils Setup File for the mx Extensions BASE distribution.

"""
#
# Load configuration(s)
#
import egenix_mx_base
configurations = (egenix_mx_base,)

#
# Run distutils setup...
#
import mxSetup
mxSetup.run_setup(configurations)



""" Tools - Add-ons for Python written in C for performance.

    Copyright (c) 2000, Marc-Andre Lemburg; mailto:mal@lemburg.com
    Copyright (c) 2000-2013, eGenix.com Software GmbH; mailto:info@egenix.com
    See the documentation for further information on copyrights,
    or contact the author. All Rights Reserved.

"""#"

# Import C extensions' symbols
from mxTools import *
from mxTools import __version__

# Needed Python imports
import operator,types,time,sys,os,re

#############################################################################
#
# Experimental function prototypes written in Python
#

def sortedby(sequence,*indices):

    """ sortedby(sequence,*indices)
    
        Returns a list representing the sequence sorted ascending by
        the fields pointed to by the additional arguments
        (indices). sequence must be at least two-dimensional, e.g. a
        list of tuples.

    """
    if len(sequence) == 0:
        return []
    x = apply(tuples,tuple(extract(lists(sequence),indices))+(sequence,))
    x.sort()
    return map(get,x,(-1,)*len(x))

def projection(sequence,*indices):

    """ projection(sequence,*indices)

        Experimental function that extracts columns from tables
        (sequence of sequences). If only one index is given, a list of
        all elements in that dimension is returned. For more indices,
        the list will contain tuples with entries for each given
        dimension.

    """
    if len(sequence) == 0:
        return []
    if len(indices) == 1:
        return lists(sequence)[indices[0]]
    else:
        return tuples(extract(lists(sequence),indices))

def frange(x,y,ticks):

    """frange(x,y,ticks)

       Returns a list of ticks equidistant floating point values from
       the interval [x,y] such that the first is equal to x and the
       last equal to y.

    """
    l = [x] * ticks
    fticks = float(ticks-1)
    diff = y - x
    for i,value in irange(l):
        l[i] = value + diff*(i/fticks)
    return l


def issequence(obj,

               isSequenceType=operator.isSequenceType,
               InstanceType=types.InstanceType):

    """issequence(obj)

       Returns 1 iff obj defines the sequence protocol, o
       otherwise. For instances at least __getitem__ must be defined.
    
    """
    rc = isSequenceType(obj)
    if rc and type(obj) == InstanceType:
        rc = hasattr(obj,'__getitem__')
    return rc

def defined(name):

    """ defined(name)

        Return 1/0 depending on whether name is a defined symbol
        in the caller's namespace.

    """
    frame = sys.cur_frame(1)
    # Look up the symbol name
    ok = frame.f_locals.has_key(name) or \
         frame.f_globals.has_key(name) or \
         frame.f_builtins.has_key(name)
    del frame
    return ok

def acqchain(obj):

    """ acqchain(obj)

        Returns a list of object representing the acquisition
        chain that the new builtin acquire() would scan.

        The order is top to bottom, with obj always being the
        last entry in the list.

    """
    l = []
    append = l.append
    while obj:
        append(obj)
        obj = obj.baseobj
    l.reverse()
    return l

# Truth constants
True = (1==1)
False = (1==0)

def reval(codestring,locals=None,

          eval=eval):

    """ Restricted execution eval().

        After a suggestion by Tim Peters on comp.lang.python.  locals
        can be given as local namespace to use when evaluating the
        codestring.

    """
    if locals is not None:
        return eval(codestring,{'__builtins__':{}},locals)
    else:
        return eval(codestring,{'__builtins__':{}})

def docstring():

    """ Returns the doc string of the calling function.

        Note that this only works for Python functions since it relies
        on the code object of the calling function.

    """
    return cur_frame(1).f_code.co_consts[0]

# Aliases for some of the APIs
nonzero = truth

# XXX This should probably be moved to mx.TextTools...

_hexcode = ("00", "01", "02", "03", "04", "05", "06", "07", "08", "09",
"0a", "0b", "0c", "0d", "0e", "0f", "10", "11", "12", "13", "14",
"15", "16", "17", "18", "19", "1a", "1b", "1c", "1d", "1e", "1f",
"20", "21", "22", "23", "24", "25", "26", "27", "28", "29", "2a",
"2b", "2c", "2d", "2e", "2f", "30", "31", "32", "33", "34", "35",
"36", "37", "38", "39", "3a", "3b", "3c", "3d", "3e", "3f", "40",
"41", "42", "43", "44", "45", "46", "47", "48", "49", "4a", "4b",
"4c", "4d", "4e", "4f", "50", "51", "52", "53", "54", "55", "56",
"57", "58", "59", "5a", "5b", "5c", "5d", "5e", "5f", "60", "61",
"62", "63", "64", "65", "66", "67", "68", "69", "6a", "6b", "6c",
"6d", "6e", "6f", "70", "71", "72", "73", "74", "75", "76", "77",
"78", "79", "7a", "7b", "7c", "7d", "7e", "7f", "80", "81", "82",
"83", "84", "85", "86", "87", "88", "89", "8a", "8b", "8c", "8d",
"8e", "8f", "90", "91", "92", "93", "94", "95", "96", "97", "98",
"99", "9a", "9b", "9c", "9d", "9e", "9f", "a0", "a1", "a2", "a3",
"a4", "a5", "a6", "a7", "a8", "a9", "aa", "ab", "ac", "ad", "ae",
"af", "b0", "b1", "b2", "b3", "b4", "b5", "b6", "b7", "b8", "b9",
"ba", "bb", "bc", "bd", "be", "bf", "c0", "c1", "c2", "c3", "c4",
"c5", "c6", "c7", "c8", "c9", "ca", "cb", "cc", "cd", "ce", "cf",
"d0", "d1", "d2", "d3", "d4", "d5", "d6", "d7", "d8", "d9", "da",
"db", "dc", "dd", "de", "df", "e0", "e1", "e2", "e3", "e4", "e5",
"e6", "e7", "e8", "e9", "ea", "eb", "ec", "ed", "ee", "ef", "f0",
"f1", "f2", "f3", "f4", "f5", "f6", "f7", "f8", "f9", "fa", "fb",
"fc", "fd", "fe", "ff")

def hexencode(data,

              hexcode=_hexcode,ord=ord):

    """ HEX encode a data string.

        Encoding is done character per character using two byte
        lower-case HEX characters.

    """
    l = []
    append = l.append
    for c in data:
        append(hexcode[ord(c)])
    return ''.join(l)


#############################################################################
#
# Utilities written in Python
#

def scanfiles(files, dir=None, levels=0, filefilter=None,

              filedict=None,join=os.path.join,isdir=os.path.isdir,
              listdir=os.listdir):

    """ Build a list of filenames starting with the filenames and
        directories given in files.

        The filenames in are made absolute relative to dir. dir
        defaults to the current working directory if not given.

        If levels is greater than 0, directories in the files list are
        recursed into up the given number of levels.

        If filefilter is given, as re match object, then all filenames
        (the absolute names) are matched against it. Filenames which
        do not match the criteria are removed from the list.

        Note that directories are not included in the resulting list.
        All filenames are non-directories.

    """
    if not files:
        return files

    # Make file names absolute and eliminate duplicates
    if dir is None:
        dir = os.getcwd()
    if filedict is None:
        filedict = {}
        recursing = 0
    else:
        recursing = 1
    dirs = []
    for file in files:
        abspath = join(dir, file)
        if isdir(abspath):
            dirs.append(abspath)
        elif filefilter is not None and \
             filefilter.match(abspath) is None:
            continue
        else:
            filedict[abspath] = 1

    # Recurse into subdirs
    if levels > 0:
        for dir in dirs:
            scanfiles(listdir(dir), dir, levels+1, filefilter, filedict)

    # Fast path: don't return file list inside recursion
    if not recursing:
        return filedict.keys()

class DictScan:

    """ Forward iterator for Python dictionaries.

        Note that no precaution is taken to insure that the dictionary
        is not modified in between calls to the __getitem__ method. It
        is the user's responsibility to ensure that the dictionary is
        neither modified, nor changed in size, since this would result
        in skipping entries or double occurance of items in the scan.

        The iterator inherits all methods from the underlying
        dictionary for convenience.

    """
    def __init__(self,dictionary):

        self.dictionary = dictionary
        self.position = 0

    def reset(self):

        """ Resets the iterator to its initial position.
        """
        self.position = 0

    def __getitem__(self,index,

                    dictscan=dictscan):

        """ "for x in iterator" interface.

            Note: for loops are cancelled by raising an IndexError.

        """
        # This may raise an IndexError which we *don't* catch
        # on purpose
        k,v,self.position = dictscan(self.dictionary,self.position)
        return k,v

    def __getattr__(self,name,

                    getattr=getattr):

        """ Inherit all other methods from the underlying dictionary.
        """
        return getattr(self.dictionary,name)

# Alias
DictItems = DictScan

_integerRE = re.compile('\s*(-?\d+)\s*$')
_integerRangeRE = re.compile('\s*(-?\d+)\s*-\s*(-?\d+)\s*$')

def srange(s,

           integer=_integerRE,
           integerRange=_integerRangeRE):

    """ Converts a textual representation of integer numbers and ranges
        to a Python list.

        Supported formats: 2,3,4,2-10,-1 - -3, 5 - -2

        Values are appended to the created list in the order specified
        in the string.

    """
    l = []
    append = l.append
    for entry in s.split(','):
        m = integer.match(entry)
        if m:
            append(int(m.groups()[0]))
            continue
        m = integerRange.match(entry)
        if m:
            start,end = map(int,m.groups())
            l[len(l):] = range(start,end+1)
    return l

def fqhostname(hostname=None, ip=None):

    """ Tries to return the fully qualified (hostname, ip) for the
        given hostname.

        If hostname is None, the default name of the local host is
        chosen. ip then defaults to '127.0.0.1' if not given.

        The function modifies the input data according to what it
        finds using the socket module. If that doesn't work the input
        data is returned unchanged.

    """
    try:
        import socket
    except ImportError:
        if hostname is None:
            hostname = os.environ.get('HOSTNAME',None)
        if ip is None:
            ip = '127.0.0.1'
        return hostname, ip
    try:
        if hostname is None:
            if ip is None:
                ip = '127.0.0.1'
            hostname = socket.gethostname()
        ip = socket.gethostbyname(hostname)
        hostname = socket.gethostbyaddr(ip)[0]
    except socket.error:
        pass
    return hostname,ip

def splitdomain(hostname=None):

    """ Tries to determine the domain name of the given hostname and
        returns a tuple (host, domain).

        If hostname is not given, the default name of the local host
        is chosen as reference.

    """
    hostname, ip = fqhostname(hostname)
    l = hostname.split('.', 1)
    if len(l) == 1:
        return (hostname, '')
    return tuple(l)

def username(default=''):

    """ Return the user name of the user running the current process.

        If no user name can be determined, default is returned.

    """
    import getpass
    try:
        return getpass.getuser()
    except:
        return default

######################################################################
#
# Old lib/Tools.py module...
#
# XXX Some of these functions are obsolete.
#

def tb_lineno(tb):

    """ Calculate the correct line number of the
        traceback given in tb (even with -O on)
    """
    c = tb.tb_frame.f_code
    tab = c.co_lnotab
    line = c.co_firstlineno
    stopat = tb.tb_lasti
    addr = 0
    for i in range(0,len(tab),2):
        addr = addr + ord(tab[i])
        if addr > stopat:
            break
        line = line + ord(tab[i+1])
    return line

def execpyc(filename,globals=None,locals=None):

    """ Execute a byte compiled file filename in globals, locals 
    """
    import marshal
    f = open(filename,'rb')
    f.read(8) # skip header (id check omitted)
    code = marshal.load(f)
    exec code in globals,locals
    
def loadpyc(filename):

    """ Load the code from a byte compiled file filename and return it
        as code object.
    """
    import marshal
    f = open(filename,'rb')
    f.read(8) # skip header (id check omitted)
    return marshal.load(f)

def import_code(name,code):

    """ Imports a code object as module name.

        Returns the previously registered module in case the module
        name was already imported. name has to be the full package
        name (pkg.pkg.mod) for the module; package local names are not
        supported and will result in top-level modules to be created.

    """
    import imp,sys
    if sys.modules.has_key(name):
        return sys.modules[name]
    m = imp.new_module(name)
    exec code in m.__dict__
    sys.modules[name] = m
    return m

def pairs2tuples(tab):

    """ Format a sequence of adjacent pairs into a list
        of 2-tuples, e.g. 'abcdef' gives [('a','b'),('c','d'),
        ('e','f')]
    """
    l = map( None, tab[:-1], tab[1:], (1,0)*(len(tab)/2) )
    l = filter( lambda x:x[2], l)
    l = map( lambda x:x[:2], l)
    return l

def exec_frame(level=0,
               
               exc_info=sys.exc_info):

    """ Return the execution frame level positions up the
        execution stack (defaulting to the current frame).

        WARNING: Storing the frame in variables will cause
        circular references which could result in the frames
        and associated objects to live forever.
    """
    try:
        1/0
    except:
        frame = exc_info()[2].tb_frame.f_back
    if level:
        for i in trange(level):
            frame = frame.f_back
    return frame

def freeze(classobj):

    """ Add all known attributes of base classes to classobj's
        attribute dictionary
        - does not overwrite attributes
    """
    dict = classobj.__dict__
    if dict.has_key('__frozen__'):
        return
    # This won't overwrite anything, but still update the class
    # dictionary in place (frozen() returns a dictionary that includes
    # dict's entries among others):
    dict.update(frozen(classobj))
    dict['__frozen__'] = 1
    # XXX How to optimize unnecessary failing lookups in baseclasses ?

def frozen(classobj):

    """ Return a dictionary that contains all known attributes
        of classobj
        - uses cached versions if available
    """
    dict = {}
    for c in reverse(classobj.__bases__):
        otherdict = c.__dict__
        if not otherdict.has_key('__frozen__'):
            frozendict = frozen(c)
        else:
            frozendict = otherdict
        dict.update(frozendict)
    dict.update(classobj.__dict__)
    return dict

def attributes(obj,of_class=None,
               d=None):

    """ Find all attributes that are accessible through obj
        and return them as dictionary. 

        If of_class is given, only those attributes are returned that
        are instances of that class. The function mimics the
        inheritance scheme used by Python.

    """
    if d is None:
        d = {}
    # First the class attributes
    classobj = getattr(obj,'__class__',None)
    if classobj is not None:
        class_attributes(classobj,of_class,d)
    # Then the instance attributes
    if of_class is not None:
        for k,v in obj.__dict__.items():
            if isinstance(v,of_class):
                d[k] = v
    else:
        d.update(obj.__dict__)
    return d

def class_attributes(classobj,of_class=None,
                     d=None):

    """ Find all attributes that are accessible through classobj
        and return them as dictionary. 

        If of_class is given, only those attributes are returned that
        are instances of that class. The function mimics the
        inheritance scheme used by Python.

    """
    if d is None:
        d = {}
    # First the base classes
    bases = getattr(classobj,'__bases__',None)
    if bases is not None:
        for b in reverse(bases):
            class_attributes(b,of_class,d)
    # Then the class itself
    if of_class is not None:
        for k,v in classobj.__dict__.items():
            if isinstance(v,of_class):
                d[k] = v
    else:
        d.update(classobj.__dict__)
    return d

def inst_attributes(obj,of_class=None):

    """ Find all instance attributes of obj that are instances of
        of_class and return them as dictionary.

    """
    d = {}
    if of_class:
        for k,v in obj.__dict__.items():
            if isinstance(v,of_class):
                d[k] = v
    else:
        d.update(obj.__dict__)
    return d

def localize(instance):

    """ Add all known attributes of the instance's class and direct
        base classes to its attribute dictionary, binding methods if
        necessary
        - only one level deep
        - does not overwrite attributes

        *WARNING:* this function introduces lots of circular
        references (one for each method) !!!  Be sure to clear
        instance.__dict__ before unscoping instance !!!

    """
    classobj = instance.__class__
    classes = (classobj,) + classobj.__bases__
    dict = instance.__dict__
    for c in classes:
        for a in c.__dict__.keys():
            if not dict.has_key(a):
                dict[a] = getattr(instance,a)

def localized(instance):

    """ Add all known attributes of the instance's class and direct
        base classes to a dictionary, binding methods if necessary, and
        return it.
        - only one level deep

        *WARNING:* this function introduces lots of circular
        references !!!  Be sure to clear the returned dictionars
        before unscoping it !!!

    """
    classobj = instance.__class__
    classes = (classobj,) + classobj.__bases__
    dict = instance.__dict__.copy()
    for c in classes:
        for a in c.__dict__.keys():
            if not dict.has_key(a):
                dict[a] = getattr(instance,a)
    return dict

def ascii2int(str, base=10,

              int=int):

    """ Convert a string to an integer.

        Works like int() except that in case of an error no
        exception raised but 0 is returned; this makes it useful in
        conjunction with map().

    """
    try:
        return int(str, base)
    except:
        return 0

def str2time(x,

             time=time,ascii2int=ascii2int):

    """ Convert a textual representation of date/time into an internal
        time.time() value using some assumptions on abbreviations.
        - returns negative numbers to indictate errors
        - knows about DST (makes small errors near the time of switching)
        - centuries can be omitted: 0-69 becomes 20xx, 70-99 19xx

        Known formats:
        - 1.1.90, 01.01.90, 1.1.1990 (date only, time defaults to 0:00:00)
        - 1.1. (current year is appended, 0:00:00)
        - 1.1.90 14:00
        - 14:00 (today is used as date)
    """
    now = time.localtime(time.time())
    ti = now[3:5]
    da = now[0:3]
    try:
        x = x.split()
        if len(x) == 0: return -3
        d = x[0].split('.')
        t = x[-1].split(':')
        if len(t) > 1:
            # Date and time
            if len(t) < 2: return -1
            ti = map(ascii2int,t)
            if len(ti) == 2:
                ti.append(0)
            elif len(ti) > 3:
                ti = ti[:3]
            ti = tuple(ti)
        else:
            # No time given, presume 0:00:00
            ti = (0,0,0)
        if len(d) > 1:
            # Date given
            if len(d) != 3: return -2
            d.reverse()
            da = tuple([int(value) for value in d])
            if da[0] == 0:
                # Year is missing
                da = (now[0],) + da[1:3]
            elif da[0] < 100:
                # Century is missing (note: this is bad !)
                if da[0] < 70: # XXX
                    da = (2000+da[0],) + da[1:3]
                else:
                    da = (1900+da[0],) + da[1:3]
        if 1 == len(d) == len(t): return -3
        try:
            tm = time.mktime(da+ti+(0,0,0))
        except:
            #sys.stderr.write('Wrong date: %s was converted to %s'%(x,`da+ti+(0,0,0,0)`))
            return -4
        if time.localtime(tm)[8] == 1:
            # DST is on, adjust time
            tm = tm - 3600
        return tm
    except:
        return -9

def filecontent(filename,default=''):

    """ Return the file's content as a string, default in case
        there's an error
    """
    try:
        f = open(filename,'rb')
    except IOError:
        return default
    c = f.read()
    f.close()
    return c

def long2str(x):

    """ Convert long integer x to a bytes string.
    """
    l = ( x        & 0xff,
         (x >> 8)  & 0xff,
         (x >> 16) & 0xff,
         (x >> 24) & 0xff)
    return ''.join([chr(x) for x in l])
    
### Hack to enable module finalization

class ModuleFinalization:

    def __init__(self,function):

        self.fini = function

    def __del__(self):

        self.fini()

# example:
#def _cleanup():
#    print 'away we go...'
#_fini = ModuleFinalization(_cleanup)

def func_info(level=1):

    """ func_info()

        Returns a tuple (name,filename) giving the name of the calling
        function (*) and the filename where it is defined.

        Note that this only works if the calling function is a Python
        function or method (because only these create new execution
        frames). When called from e.g. a builtin function like map(),
        it will return information about the function from where the
        builtin function was called.

        (*) level indicates how far up the calling stack to look for
        the information. Default is one level meaning: the calling
        function.
        
    """
    try:
        1/0
    except:
        frame = sys.exc_info()[2].tb_frame
        for i in trange(level):
            frame = frame.f_back
        name = frame.f_code.co_name
        filename = frame.f_code.co_filename
        del frame # you never know...
    return (name,filename)

def func_sig(func):

    """func_sig(func)

       Returns the signature of a Python function/method as string.
       Keyword initializers are also shown using
       repr(). Representations longer than 100 bytes are truncated.

       XXX Anonymous argument ((a,b,c)=(1,2,3)) are not supported and
           probably never will be since they require disassembling the
           byte code which is bound to fail once byte code optimizers find
           their way into every Pythoneers home...

    """
    if hasattr(func,'im_func'):
        # func is a method
        func = func.im_func
    code = func.func_code
    fname = code.co_name
    callargs = code.co_argcount
    # XXX Uses hard coded values taken from Include/compile.h
    args = list(code.co_varnames[:callargs])
    if func.func_defaults:
        i = len(args) - len(func.func_defaults)
        for default in func.func_defaults:
            try:
                r = repr(default)
            except:
                r = '<repr-error>'
            if len(r) > 100:
                r = r[:100] + '...'
            arg = args[i]
            if arg[0] == '.':
                # anonymous arguments
                arg = '(...)'
            args[i] = '%s=%s' % (arg,r)
            i = i + 1
    if code.co_flags & 0x0004: # CO_VARARGS
        args.append('*'+code.co_varnames[callargs])
        callargs = callargs + 1
    if code.co_flags & 0x0008: # CO_VARKEYWORDS
        args.append('**'+code.co_varnames[callargs])
        callargs = callargs + 1
    return '%s(%s)' % (fname,', '.join(args))

def func_call(level=1):

    """ func_call()

        Returns a string explaining which parameters where passed to
        the calling function (*) and from which file and line number it
        was invoked.  Same comments as for func_info(). Note that
        line number information is only correct when running Python
        in non-optimized mode (i.e. without -O). Sample return string:

        'test(a=1, b=2, c=3, args=()) # called from "Tools.py":353'

        (*) level indicates how far up the calling stack to look for
        the information. Default is one level meaning: the calling
        function.
        
    """
    try:
        1/0
    except:
        frame = sys.exc_info()[2].tb_frame
    for i in trange(level):
        frame = frame.f_back
    #import hack; hack.show(frame,5)
    code = frame.f_code
    fname = code.co_name
    l = []
    callargs = code.co_argcount
    # XXX Uses hard coded values taken from Include/compile.h
    if code.co_flags & 0x0004: # CO_VARARGS
        callargs = callargs + 1
    if code.co_flags & 0x0008: # CO_VARKEYWORDS
        callargs = callargs + 1
    for v in code.co_varnames[:callargs]:
        try:
            r = repr(frame.f_locals[v])
        except:
            r = '<repr-error>'
        if len(r) > 100:
            r = r[:100] + '...'
        l.append('%s=%s' % (v,r))
    if frame.f_back:
        where = '# called from "%s":%i' % \
                (frame.f_back.f_code.co_filename,frame.f_back.f_lineno)
    else:
        where = '# called from <toplevel>'
    del frame,code # you never know...
    return '%s(%s) %s' % (fname,', '.join(l), where)

def localize_builtins():

    """ Copy all builtins to the caller's locals. This is done
        in a non-overwriting fashion.
    """
    try:
        1/0
    except:
        frame = sys.exc_info()[2].tb_frame.f_back
    builtins = frame.f_builtins
    locals = frame.f_locals
    for k,v in builtins.items():
        if not locals.has_key(k):
            locals[k] = v
    del frame,builtins,locals # better safe than sorry

_basemethod_cache = {}

def basemethod(object,method=None,

               cache = _basemethod_cache,InstanceType=types.InstanceType,
               ClassType=types.ClassType):

    """ Return the unbound method that is defined *after* method in the
        inheritance order of object with the same name as method
        (usually called base method or overridden method).

        object can be an instance, class or bound method. method, if
        given, may be a bound or unbound method. If it is not given,
        object must be bound method.

        Note: Unbound methods must be called with an instance as first
        argument.

        The function uses a cache to speed up processing. Changes done
        to the class structure after the first hit will not be noticed
        by the function.

        XXX Rewrite in C to increase performance.

    """
    if method is None:
        method = object
        object = method.im_self
    defclass = method.im_class
    name = method.__name__
    if type(object) is InstanceType:
        objclass = object.__class__
    elif type(object) is ClassType:
        objclass = object
    else:
        objclass = object.im_class

    # Check cache
    cacheentry = (objclass, name)
    basemethod = cache.get(cacheentry, None)
    if basemethod is not None:
        if not issubclass(objclass, basemethod.im_class):
            if __debug__:
                sys.stderr.write(
                    'basemethod(%s, %s): cached version (%s) mismatch: '
                    '%s !-> %s\n' %
                    (object, method, basemethod,
                     objclass, basemethod.im_class))
        else:
            return basemethod

    # Find defining class
    path = [objclass]
    while 1:
        if not path:
            raise AttributeError,method
        c = path[0]
        del path[0]
        if c.__bases__:
            # Prepend bases of the class
            path[0:0] = list(c.__bases__)
        if c is defclass:
            # Found (first occurance of) defining class in inheritance
            # graph
            break
        
    # Scan rest of path for the next occurance of a method with the
    # same name
    while 1:
        if not path:
            raise AttributeError,name
        c = path[0]
        basemethod = getattr(c, name, None)
        if basemethod is not None:
            # Found; store in cache and return
            cache[cacheentry] = basemethod
            return basemethod
        del path[0]
    raise AttributeError,'method %s' % name
    
def lookup_path(classobj):

    """ Return a list representing the lookup path taken by getattr()
        whenever an attribute from classobj is requested.

        The path consists of all class objects passed during lookup in
        the right order.
        
    """
    path = [classobj]
    for i,c in reverse(irange(path)):
        if c.__bases__:
            l = []
            for bc in c.__bases__:
                l[len(l):] = lookup_path(bc)
            path[i+1:i+1] = l
    return path

hexcode = tuple('0123456789abcdefghijklmnopqrstuvwxyz')
code64 = tuple('ABCDEFGHIJKLMNOPQRSTUVWXYZ'
               'abcdefghijklmnopqrstuvwxyz'
               '0123456789+/')
code256 = tuple(map(chr,range(256)))

def base(x,b,code=hexcode,
         # Locals:
         divmod=divmod):
    
    """ Return a string representation of integer x in base b.

        Uses code as encoding table (defaults to an extended HEX
        table).
        
    """
    if x == 0: return '0'
    l = []
    append = l.append
    while x > 0:
        x,y = divmod(x,b)
        append(code[y])
    l.reverse()
    return ''.join(l)

def filedate(path,

             stat=os.stat):

    """ Return the modification date/time as DateTime instance.

        Needs mxDateTime to be installed.

    """
    from mx import DateTime
    return DateTime.localtime(stat(path)[8])

def filesize(path,

             stat=os.stat):

    """ Return the file size in bytes
    """
    return stat(path)[6]

def abspath(path,

            expandvars=os.path.expandvars,expanduser=os.path.expanduser,
            join=os.path.join,getcwd=os.getcwd):

    """ Return the corresponding absolute path for path.

        path is expanded in the usual shell ways before
        joining it with the current working directory.

    """
    try:
        path = expandvars(path)
    except AttributeError:
        pass
    try:
        path = expanduser(path)
    except AttributeError:
        pass
    return join(getcwd(),path)

def _addlinebreaks(data, column, linebreak='\012'):

    """ Break data into multiple lines at column.

        Uses linebreak as end-of-line indicator (defaults to LF).
        
    """
    from cStringIO import StringIO
    infile = StringIO(data)
    outfile = StringIO()
    read = infile.read
    write = outfile.write
    chunk = read(column)
    while chunk:
        write(chunk)
        write(linebreak)
        chunk = read(column)
    return outfile.getvalue()

def _uu_encode(data, filename='<data>', mode=0666):
    
    from cStringIO import StringIO
    from binascii import b2a_uu
    infile = StringIO(data)
    outfile = StringIO()
    read = infile.read
    write = outfile.write
    # Encode
    write('begin %o %s\n' % (mode & 0777, filename))
    chunk = read(45)
    while chunk:
        write(b2a_uu(chunk))
        chunk = read(45)
    write(' \nend\n')
    return outfile.getvalue()

def _uu_decode(data):
    
    from cStringIO import StringIO
    from binascii import a2b_uu
    infile = StringIO(input)
    outfile = StringIO()
    readline = infile.readline
    write = outfile.write

    # Find start of encoded data
    while 1:
        s = readline()
        #print '...',s,
        if not s:
            raise ValueError, 'Missing "begin" line in input data'
        if s[:5] == 'begin':
            break

    # Decode
    while 1:
        s = readline()
        if not s or \
           s == 'end\n':
            break
        try:
            data = a2b_uu(s)
        except binascii.Error, v:
            # Workaround for broken uuencoders by /Fredrik Lundh
            nbytes = (((ord(s[0])-32) & 63) * 4 + 5) / 3
            data = a2b_uu(s[:nbytes])
            #sys.stderr.write("Warning: %s\n" % str(v))
        write(data)
    if not s:
        raise ValueError, 'Truncated input data'

    return outfile.getvalue()

def encodedata(data, encoding):

    """ Encode data using the given encoding.

        Possible values for encoding include:
        'base64'   - BASE 64 encoding
        'hex'      - HEX encoding (no line breaks)
        'hexlines' - HEX encoding (with CR line breaks)

        In Python 2.0 and up, encoding may also be an encoding
        supported natively by Python via the codec registry.
    
    """
    encoding = encoding.lower()
    
    if encoding == 'base64':
        import base64
        return base64.encodestring(data)
    
    elif encoding == 'hex' or \
         encoding == 'hexlines':
        from mx.TextTools import str2hex
        result = str2hex(data)
        if encoding == 'hexlines':
            return _addlinebreaks(result, 72)
        return result

    else:
        # This works in Python >=2.0 only
        try:
            return data.encode(encoding)
        except AttributeError:
            raise ValueError, 'unknown encoding "%s"' % encoding

def decodedata(data, encoding):

    """ Decode data using the given encoding.

        Possible values for encoding include:
        'base64'   - BASE 64 encoding
        'hex'      - HEX encoding (no line breaks)
        'hexlines' - HEX encoding (with CR line breaks)

        In Python 2.0 and up, encoding may also be an encoding
        supported natively by Python via the codec registry.
    
    """
    encoding = encoding.lower()
    
    if encoding == 'base64':
        import base64
        return base64.decodestring(data)
    
    elif encoding == 'hex' or \
         encoding == 'hexlines':
        from mx.TextTools import hex2str
        # Remove whitespace
        l = data.split()
        data = ''.join(l)
        # Decode
        return hex2str(data)

    else:
        # This works in Python >=2.0 only
        try:
            from codecs import lookup
        except ImportError:
            raise ValueError, 'unknown encoding "%s"' % encoding
        else:
            decode = lookup(encoding)[1]
            return decode(data)



# -*- coding: latin-1 -*-

from mx.TextTools import __version__
from mx.TextTools.Examples.HTML import *
from mx.TextTools.Constants.TagTables import *
import pprint, time, pickle

print 'Testing mxTextTools version', __version__
print

# Test for Unicode
try:
    unicode
except NameError:
    HAVE_UNICODE = 0
else:
    HAVE_UNICODE = 1
    ua = unicode('a')
    ub = unicode('b')
    uc = unicode('c')
    ud = unicode('d')
    ue = unicode('e')
    uabc = unicode('abc')
    uHello = unicode('Hello')
    uempty = unicode('')

# Find a HTML file usable for the test
if len(sys.argv) > 1:
    filenames = sys.argv[1:]
else:
    filenames = ['/usr/share/doc/packages/mysql/html/manual.html',
                 '../Doc/mxTextTools.html']
text = ''
for filename in filenames:
    try:
        text = open(filename).read()
    except IOError:
        pass
    else:
        print 'HTML file used for testing the Tagging Engine:'
        print '  ', filename
        print
        break
if not text:
    text = """
<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01 Transitional//EN">
<HTML>
<HEAD>
   <TITLE>mx Extension Series - License Information</TITLE>
   <META HTTP-EQUIV="Content-Type" CONTENT="text/html; charset=iso-8859-1">
</HEAD>

  <BODY TEXT="#000000" BGCOLOR="#FFFFFF" LINK="#0000EE" VLINK="#551A8B" ALINK="#FF0000">
...
    <CENTER><FONT SIZE=-1>&copy; 2000, Copyright by eGenix.com
    Software GmbH, Langengeld, Germany; All Rights Reserved. mailto:
    <A HREF="mailto:info@egenix.com">info@egenix.com</A>
    </FONT></CENTER>
  </BODY>
</HTML>

"""#"

# Test suite
while 1:

    if 1:
        print 'Tagging Engine:'
        print ' parsing HTML ...',
        utext = upper(text)
        t = time.clock()
        result, taglist, nextindex = tag(utext, htmltable)
        assert result == 1
        print ' done. (%5.2f sec.; len(taglist)=%i)' % \
              (time.clock() - t, len(taglist))
        if HAVE_UNICODE: 
            print ' parsing Unicode HTML ...',
            try:
                uutext = unicode(utext, 'latin-1')
            except UnicodeError:
                print ' ... HTML data must be Latin-1; skipping test.'
            else:
                t = time.clock()
                result, utaglist, nextindex = tag(uutext, htmltable)
                assert result == 1
                print ' done. (%5.2f sec.; len(utaglist)=%i)' % \
                      (time.clock() - t, len(utaglist))
                assert taglist == utaglist
                utaglist = None
                uutext = None
        utext = None
        taglist = None

        print ' testing some tag table semantics...'
        table = ((None,Word,'Word'),)
        assert tag('Word',table)[0] == 1
        assert tag('word',table)[0] == 0
        assert tag('xyz',table)[0] == 0

        table = ((None,Word,'Word',MatchFail),)
        assert tag('Word',table)[0] == 1
        assert tag('word',table)[0] == 0
        assert tag('xyz',table)[0] == 0

        table = ((None,Word,'Word',MatchOk),)
        assert tag('Word',table)[0] == 1
        assert tag('word',table)[0] == 1
        assert tag('xyz',table)[0] == 1

        table = ((None,Word,'Word',MatchOk,MatchFail),)
        assert tag('Word',table)[0] == 0
        assert tag('word',table)[0] == 1
        assert tag('xyz',table)[0] == 1

        print ' done.'

        #continue

    print 'splitat()'
    assert splitat('Hello','l') == ('He', 'lo')
    assert splitat('Hello','l',2) == ('Hel', 'o')
    assert splitat('Hello','l',-1) == ('Hel', 'o')
    assert splitat('Hello','l',-2) == ('He', 'lo')
    if HAVE_UNICODE:
        assert splitat(uHello,'l') == (unicode('He'), unicode('lo'))
        assert splitat(uHello,'l',2) == (unicode('Hel'), unicode('o'))
        assert splitat(uHello,unicode('l'),-1) == (unicode('Hel'), unicode('o'))
        assert splitat(uHello,unicode('l'),-2) == (unicode('He'), unicode('lo'))

    print 'suffix()'
    assert suffix('abc.html/',('.htm','abc','.html','/'),0,3) == 'abc'
    assert suffix('abc.html/',('.htm','abc','.html','/'),0,4) == None
    assert suffix('abc.html/',('.htm','abc','.html','/'),0,8) == '.html'
    if HAVE_UNICODE:
        assert suffix(unicode('abc.html/'),('.htm','abc','.html','/'),0,3) == unicode('abc')
        assert suffix(unicode('abc.html/'),(unicode('.htm'),unicode('abc'),'.html','/'),0,3) == unicode('abc')
        assert suffix(unicode('abc.html/'),('.htm','abc',unicode('.html'),'/'),0,3) == unicode('abc')
        try:
            suffix('abc.html/',(unicode('.htm'),'abc','.html','/'))
        except TypeError:
            pass
        else:
            raise AssertionError, \
                  'suffix(string,...) should not accept unicode suffixes'
        assert suffix(unicode('abc.html/'),('.htm','abc','.html','/'),0,4) == None

    print 'prefix()'
    assert prefix('abc.html/',('.htm','abc','.html','/'),0,3) == 'abc'
    assert prefix('abc.html/',('.htm','abc','.html','/'),1,4) == None
    assert prefix('abc.html/',('.htm','abc','.html','/'),3,9) == '.htm'
    if HAVE_UNICODE:
        assert prefix(unicode('abc.html/'),('.htm','abc','.html','/'),0,3) == unicode('abc')
        assert prefix(unicode('abc.html/'),(unicode('.htm'),unicode('abc'),'.html','/'),0,3) == unicode('abc')
        assert prefix(unicode('abc.html/'),('.htm','abc',unicode('.html'),'/'),0,3) == unicode('abc')
        try:
            prefix('abc.html/',(unicode('.htm'),'abc','.html','/'))
        except TypeError:
            pass
        else:
            raise AssertionError, \
                  'prefix(string,...) should not accept unicode prefixes'
        assert prefix(unicode('abc.html/'),('.htm','abc','.html','/'),0,4) == unicode('abc')

    print 'join()'
    assert join(('a','b','c')) == 'abc'
    assert join(['a','b','c']) == 'abc'
    assert join(('a','b','c'),' ') == 'a b c'
    assert join(['a','b','c'],' ') == 'a b c'
    assert join((('abc',0,1),('abc',1,2),('abc',2,3))) == 'abc'
    assert join((('abc',0,1),'b',('abc',2,3))) == 'abc'
    assert join((('abc',0,3),)) == 'abc'
    if HAVE_UNICODE:
        assert join((ua,ub,uc)) == uabc
        assert join([ua,ub,uc]) == uabc
        assert join((ua,ub,uc),' ') == unicode('a b c')
        assert join([ua,ub,uc],' ') == unicode('a b c')
        assert join(((uabc,0,1),(uabc,1,2),(uabc,2,3))) == uabc
        assert join(((uabc,0,1),ub,(uabc,2,3))) == uabc
        assert join(((uabc,0,3),)) == uabc

    print 'upper()'
    assert upper('HeLLo') == 'HELLO'
    assert upper('hello') == 'HELLO'
    assert upper('HELLO') == 'HELLO'
    assert upper('HELLO ') == 'HELLO '
    assert upper('HELLO 123') == 'HELLO 123'
    if HAVE_UNICODE:
        assert upper(unicode('HeLLo')) == unicode('HELLO')
        assert upper(unicode('hello')) == unicode('HELLO')
        assert upper(unicode('HELLO')) == unicode('HELLO')
        assert upper(unicode('HELLO ')) == unicode('HELLO ')
        assert upper(unicode('HELLO 123')) == unicode('HELLO 123')

    print 'lower()'
    assert lower('HeLLo') == 'hello'
    assert lower('hello') == 'hello'
    assert lower('HELLO') == 'hello'
    assert lower('HELLO ') == 'hello '
    assert lower('HELLO 123') == 'hello 123'
    if HAVE_UNICODE:
        assert lower(unicode('HeLLo')) == unicode('hello')
        assert lower(unicode('hello')) == unicode('hello')
        assert lower(unicode('HELLO')) == unicode('hello')
        assert lower(unicode('HELLO ')) == unicode('hello ')
        assert lower(unicode('HELLO 123')) == unicode('hello 123')

    print 'isascii()'
    assert isascii('abc') == 1
    assert isascii('abcäöü') == 0
    assert isascii('abcäöüdef') == 0
    assert isascii('.,- 1234') == 1
    if HAVE_UNICODE:
        assert isascii(uabc) == 1
        assert isascii(unicode('abcäöü', 'latin-1')) == 0
        assert isascii(unicode('abcäöüdef', 'latin-1')) == 0
        assert isascii(unicode('.,- 1234')) == 1

    print 'setstrip()'
    assert setstrip('Hello', set('')) == 'Hello'
    assert setstrip('Hello', set('o')) == 'Hell'
    assert setstrip(' Hello ', set(' o')) == 'Hell'
    assert setstrip(' Hello ', set(' o'), 0, len(' Hello '), -1) == 'Hello '
    assert setstrip(' Hello ', set(' o'), 0, len(' Hello '), 1) == ' Hell'
    assert setstrip('  ', set(' ')) == ''

    print 'setsplit()'
    assert setsplit('Hello', set('l')) == ['He', 'o']
    assert setsplit('Hello', set('lo')) == ['He',]
    assert setsplit('Hello', set('abc')) == ['Hello',]

    print 'setsplitx()'
    assert setsplitx('Hello', set('l')) == ['He', 'll', 'o']
    assert setsplitx('Hello', set('lo')) == ['He', 'llo']
    assert setsplitx('Hello', set('abc')) == ['Hello',]

    print 'joinlist()'
    assert joinlist('Hello', [('A',1,2), ('B',3,4)]) == \
           [('Hello', 0, 1), 'A', ('Hello', 2, 3), 'B', ('Hello', 4, 5)]
    assert join(joinlist('Hello', [('A',1,2), ('B',3,4)])) == \
           'HAlBo'
    if HAVE_UNICODE:
        assert joinlist(uHello, [('A',1,2), ('B',3,4)]) == \
               [(uHello, 0, 1), 'A', (uHello, 2, 3), 'B', (uHello, 4, 5)]
        assert join(joinlist(uHello, [('A',1,2), ('B',3,4)])) == \
               unicode('HAlBo')
        assert join(joinlist('Hello', [(ua,1,2), (ub,3,4)])) == \
               unicode('Halbo')

    print 'charsplit()'
    assert charsplit('Hello', 'l') == ['He', '', 'o']
    assert charsplit('Hello', 'e') == ['H', 'llo']
    assert charsplit('HelloHello', 'e') == ['H', 'lloH', 'llo']
    if HAVE_UNICODE:
        assert charsplit(uHello, unicode('l')) == [unicode('He'), unicode(''), unicode('o')]
        assert charsplit(uHello, unicode('e')) == [unicode('H'), unicode('llo')]
        assert charsplit(uHello*2, unicode('e')) == [unicode('H'), unicode('lloH'), unicode('llo')]

    print 'CharSet().contains()'
    tests = [
        ("a-z",
         ('a', 1), ('b', 1), ('c', 1), ('z', 1),
         ('A', 0), ('B', 0), ('C', 0), ('Z', 0),
         ),
        ("a\-z",
         ('a', 1), ('b', 0), ('c', 0), ('z', 1), ('-', 1),
         ),
        ]
    if HAVE_UNICODE:
        tests[len(tests):] = [
        ("a-z",
         ('a', 1), ('b', 1), ('c', 1), ('z', 1),
         ('A', 0), ('B', 0), ('C', 0), ('Z', 0),
         (unicode('a'), 1), (unicode('b'), 1), (unicode('c'), 1), (unicode('z'), 1),
         (unicode('A'), 0), (unicode('B'), 0), (unicode('C'), 0), (unicode('Z'), 0),
         ),
        ("abc",
         ('a', 1), ('b', 1), ('c', 1), ('z', 0),
         ('A', 0), ('B', 0), ('C', 0), ('Z', 0),
         (unicode('a'), 1), (unicode('b'), 1), (unicode('c'), 1), (unicode('z'), 0),
         (unicode('A'), 0), (unicode('B'), 0), (unicode('C'), 0), (unicode('Z'), 0),
         ),
        (unicode("abc"),
         ('a', 1), ('b', 1), ('c', 1), ('z', 0),
         ('A', 0), ('B', 0), ('C', 0), ('Z', 0),
         (unicode('a'), 1), (unicode('b'), 1), (unicode('c'), 1), (unicode('z'), 0),
         (unicode('A'), 0), (unicode('B'), 0), (unicode('C'), 0), (unicode('Z'), 0),
         ),
        (unicode('a-z\uFFFF', 'unicode-escape'),
         ('a', 1), ('b', 1), ('c', 1), ('z', 1),
         ('A', 0), ('B', 0), ('C', 0), ('Z', 0), 
         (unicode('a'), 1), (unicode('b'), 1), (unicode('c'), 1), (unicode('z'), 1),
         (unicode('A'), 0), (unicode('B'), 0), (unicode('C'), 0), (unicode('Z'), 0),
         (unichr(55555), 0), (unichr(1234), 0), (unichr(1010), 0),
         (unichr(0xFFFF), 1),
         ),
        (unicode("a\-z"),
         ('a', 1), ('b', 0), ('c', 0), ('z', 1), ('-', 1),
         ),
        ]

    for test in tests:
        cs = CharSet(test[0])
        for ch, rc in test[1:]:
            assert cs.contains(ch) == rc, \
                   'CharSet(%s).contains(%s) ... expected: %s' % \
                   (repr(cs.definition), repr(ch), rc)

    print 'CharSet().search()'
    tests = [
        ("a-z",
         ('', None), ('abc', 0), ('ABCd', 3),
         ),
        ("a\-z",
         ('', None), ('bcd', None), ('ABCd', None), ('zzz', 0),
         ),
        ("abc",
         ('', None), ('bcd', 0), ('ABCd', None), ('zzz', None), ('dddbbb', 3),
         ),
        ]
    if HAVE_UNICODE:
        tests[len(tests):] = [
        ("a-z",
         ('', None), ('abc', 0), ('ABCd', 3),
         (unicode(''), None), (unicode('abc'), 0), (unicode('ABCd'), 3),
         ),
        ("a\-z",
         ('', None), ('bcd', None), ('ABCd', None), ('zzz', 0),
         (unicode(''), None), (unicode('bcd'), None), (unicode('ABCd'), None), (unicode('zzz'), 0),
         ),
        ("abc",
         ('', None), ('bcd', 0), ('ABCd', None), ('zzz', None), ('dddbbb', 3),
         (unicode(''), None), (unicode('bcd'), 0), (unicode('ABCd'), None), (unicode('zzz'), None), (unicode('dddbbb'), 3),
         ),
        (unicode('a-z\uFFFF', 'unicode-escape'),
         ('', None), ('abc', 0), ('ABCd', 3),
         (unicode(''), None), (unicode('abc'), 0), (unicode('ABCd'), 3),
         (unichr(0xFFFF), 0),
         ),
        (unicode('a\-z'),
         ('', None), ('bcd', None), ('ABCd', None), ('zzz', 0),
         (unicode(''), None), (unicode('bcd'), None), (unicode('ABCd'), None), (unicode('zzz'), 0),
         ),
        (unicode('abc'),
         ('', None), ('bcd', 0), ('ABCd', None), ('zzz', None), ('dddbbb', 3),
         (unicode(''), None), (unicode('bcd'), 0), (unicode('ABCd'), None), (unicode('zzz'), None), (unicode('dddbbb'), 3),
         ),
        ]

    for test in tests:
        cs = CharSet(test[0])
        for ch, rc in test[1:]:
            assert cs.search(ch) == rc, \
                   'CharSet(%s).search(%s) ... expected: %s' % \
                   (repr(cs.definition), repr(ch), rc)

    print 'CharSet().match()'
    tests = [
        ("a-z",
         ('', 0), ('abc', 3), ('ABCd', 0),
         ),
        ("a\-z",
         ('', 0), ('bcd', 0), ('ABCd', 0), ('zzz', 3),
         ),
        ("abc",
         ('', 0), ('bcd', 2), ('ABCd', 0), ('zzz', 0), ('dddbbb', 0),
         ),
        ]
    if HAVE_UNICODE:
        tests[len(tests):] = [
        ("a-z",
         ('', 0), ('abc', 3), ('ABCd', 0),
         (unicode(''), 0), (unicode('abc'), 3), (unicode('ABCd'), 0),
         ),
        ("a\-z",
         ('', 0), ('bcd', 0), ('ABCd', 0), ('zzz', 3),
         (unicode(''), 0), (unicode('bcd'), 0), (unicode('ABCd'), 0), (unicode('zzz'), 3),
         ),
        ("abc",
         ('', 0), ('bcd', 2), ('ABCd', 0), ('zzz', 0), ('dddbbb', 0),
         (unicode(''), 0), (unicode('bcd'), 2), (unicode('ABCd'), 0), (unicode('zzz'), 0), (unicode('dddbbb'), 0),
         ),
        (unicode('a-z\uFFFF', 'unicode-escape'),
         ('', 0), ('abc', 3), ('ABCd', 0),
         (unicode(''), 0), (unicode('abc'), 3), (unicode('ABCd'), 0),
         (unichr(0xFFFF), 1),
         ),
        (unicode('a\-z'),
         ('', 0), ('bcd', 0), ('ABCd', 0), ('zzz', 3),
         (unicode(''), 0), (unicode('bcd'), 0), (unicode('ABCd'), 0), (unicode('zzz'), 3),
         ),
        (unicode('abc'),
         ('', 0), ('bcd', 2), ('ABCd', 0), ('zzz', 0), ('dddbbb', 0),
         (unicode(''), 0), (unicode('bcd'), 2), (unicode('ABCd'), 0), (unicode('zzz'), 0), (unicode('dddbbb'), 0),
         ),
        ]

    for test in tests:
        cs = CharSet(test[0])
        for ch, rc in test[1:]:
            assert cs.match(ch) == rc, \
                   'CharSet(%s).match(%s) ... expected: %s' % \
                   (repr(cs.definition), repr(ch), rc)

    print 'CharSet().strip()'
    assert CharSet('').strip('Hello') == 'Hello'
    assert CharSet('o').strip('Hello') == 'Hell'
    assert CharSet(' o').strip(' Hello ') == 'Hell'
    assert CharSet(' o').strip(' Hello ', -1, 0, len(' Hello ')) == 'Hello '
    assert CharSet(' o').strip(' Hello ', 1, 0, len(' Hello ')) == ' Hell'
    assert CharSet('  ').strip('  ') == ''
    if HAVE_UNICODE:
        assert CharSet('').strip(unicode('Hello')) == unicode('Hello')
        assert CharSet('o').strip(unicode('Hello')) == unicode('Hell')
        assert CharSet(' o').strip(unicode(' Hello ')) == unicode('Hell')
        assert CharSet(' o').strip(unicode(' Hello '), -1, 0, len(unicode(' Hello '))) == unicode('Hello ')
        assert CharSet(' o').strip(unicode(' Hello '), 1, 0, len(unicode(' Hello '))) == unicode(' Hell')
        assert CharSet(unicode('')).strip(unicode('Hello')) == unicode('Hello')
        assert CharSet(unicode('o')).strip(unicode('Hello')) == unicode('Hell')
        assert CharSet(unicode(' o')).strip(unicode(' Hello ')) == unicode('Hell')
        assert CharSet(unicode(' o')).strip(unicode(' Hello '), -1, 0, len(unicode(' Hello '))) == unicode('Hello ')
        assert CharSet(unicode(' o')).strip(unicode(' Hello '), 1, 0, len(unicode(' Hello '))) == unicode(' Hell')

    print 'CharSet().split()'
    assert CharSet('l').split('Hello') == ['He', 'o']
    assert CharSet('lo').split('Hello') == ['He',]
    assert CharSet('abc').split('Hello') == ['Hello',]
    if HAVE_UNICODE:
        assert CharSet('l').split(unicode('Hello')) == ['He', 'o']
        assert CharSet('lo').split(unicode('Hello')) == ['He',]
        assert CharSet('abc').split(unicode('Hello')) == ['Hello',]
        assert CharSet(unicode('l')).split(unicode('Hello')) == ['He', 'o']
        assert CharSet(unicode('lo')).split(unicode('Hello')) == ['He',]
        assert CharSet(unicode('abc')).split(unicode('Hello')) == ['Hello',]

    print 'CharSet().splitx()'
    assert CharSet('l').splitx('Hello') == ['He', 'll', 'o']
    assert CharSet('lo').splitx('Hello') == ['He', 'llo']
    assert CharSet('abc').splitx('Hello') == ['Hello',]
    assert CharSet(' ').splitx('x y ') == ['x', ' ', 'y', ' ']
    assert CharSet(' ').splitx(' x y ') == ['', ' ', 'x', ' ', 'y', ' ']
    if HAVE_UNICODE:
        assert CharSet('l').splitx(unicode('Hello')) == ['He', 'll', 'o']
        assert CharSet('lo').splitx(unicode('Hello')) == ['He', 'llo']
        assert CharSet('abc').splitx(unicode('Hello')) == ['Hello',]
        assert CharSet(' ').splitx(unicode('x y ')) == ['x', ' ', 'y', ' ']
        assert CharSet(' ').splitx(unicode(' x y ')) == ['', ' ', 'x', ' ', 'y', ' ']
        assert CharSet(unicode('l')).splitx(unicode('Hello')) == ['He', 'll', 'o']
        assert CharSet(unicode('lo')).splitx(unicode('Hello')) == ['He', 'llo']
        assert CharSet(unicode('abc')).splitx(unicode('Hello')) == ['Hello',]
        assert CharSet(unicode(' ')).splitx(unicode('x y ')) == ['x', ' ', 'y', ' ']
        assert CharSet(unicode(' ')).splitx(unicode(' x y ')) == ['', ' ', 'x', ' ', 'y', ' ']

    print 'CharSet() negative logic matching'
    assert CharSet('0-9').contains('a') == 0
    assert CharSet('^0-9').contains('a') == 1
    assert CharSet('0-9').match('abc') == 0
    assert CharSet('0-9').match('123abc') == 3
    assert CharSet('0-9').match('abc123') == 0
    assert CharSet('0-9').search('abc') == None
    assert CharSet('0-9').search('123abc') == 0
    assert CharSet('0-9').search('abc123') == 3
    assert CharSet('^0-9').match('abc') == 3
    assert CharSet('^0-9').match('123abc') == 0
    assert CharSet('^0-9').match('abc123') == 3
    assert CharSet('^0-9').search('abc') == 0
    assert CharSet('^0-9').search('123abc') == 3
    assert CharSet('^0-9').search('abc123') == 0
    if HAVE_UNICODE:
        assert CharSet('0-9').contains(unicode('a')) == 0
        assert CharSet('^0-9').contains(unicode('a')) == 1
        assert CharSet('0-9').match(unicode('abc')) == 0
        assert CharSet('0-9').match(unicode('123abc')) == 3
        assert CharSet('0-9').match(unicode('abc123')) == 0
        assert CharSet('0-9').search(unicode('abc')) == None
        assert CharSet('0-9').search(unicode('123abc')) == 0
        assert CharSet('0-9').search(unicode('abc123')) == 3
        assert CharSet('^0-9').match(unicode('abc')) == 3
        assert CharSet('^0-9').match(unicode('123abc')) == 0
        assert CharSet('^0-9').match(unicode('abc123')) == 3
        assert CharSet('^0-9').search(unicode('abc')) == 0
        assert CharSet('^0-9').search(unicode('123abc')) == 3
        assert CharSet('^0-9').search(unicode('abc123')) == 0
        assert CharSet(unicode('0-9')).contains(unicode('a')) == 0
        assert CharSet(unicode('^0-9')).contains(unicode('a')) == 1
        assert CharSet(unicode('0-9')).match(unicode('abc')) == 0
        assert CharSet(unicode('0-9')).match(unicode('123abc')) == 3
        assert CharSet(unicode('0-9')).match(unicode('abc123')) == 0
        assert CharSet(unicode('0-9')).search(unicode('abc')) == None
        assert CharSet(unicode('0-9')).search(unicode('123abc')) == 0
        assert CharSet(unicode('0-9')).search(unicode('abc123')) == 3
        assert CharSet(unicode('^0-9')).match(unicode('abc')) == 3
        assert CharSet(unicode('^0-9')).match(unicode('123abc')) == 0
        assert CharSet(unicode('^0-9')).match(unicode('abc123')) == 3
        assert CharSet(unicode('^0-9')).search(unicode('abc')) == 0
        assert CharSet(unicode('^0-9')).search(unicode('123abc')) == 3
        assert CharSet(unicode('^0-9')).search(unicode('abc123')) == 0

    print 'CharSet() pickling'
    cs = CharSet('abc')
    pcs = pickle.dumps(cs)
    cs1 = pickle.loads(pcs)
    assert cs.match('abc') == cs1.match('abc')
    assert cs.match('') == cs1.match('')
    assert cs.match('eee') == cs1.match('eee')
    assert cs.match('   abc') == cs1.match('   abc')
    assert cs.match('abc...d') == cs1.match('abc...d')
    assert cs.search('xxxabc') == cs1.search('xxxabc')

    ###

    htmltag = (
        (None,Is,'<'),
        # is this a closing tag ?
        ('closetag',Is,'/',+1),
        # a coment ?
        ('comment',Is,'!','check-xmp-tag'),
         (None,Word,'--',+4),
         ('text',WordStart,'-->',+1),
         (None,Skip,3),
         (None,Jump,To,MatchOk),
         # a SGML-Tag ?
         ('other',AllNotIn,'>',+1),
          (None,Is,'>'),
         (None,Jump,To,MatchOk),
        # XMP-Tag ?
        'check-xmp-tag',
        ('tagname',Word,'xmp','get-tagname'),
         (None,Is,'>'),
         ('text',WordStart,'</xmp>'),
         (None,Skip,len('</xmp>')),
         (None,Jump,To,MatchOk),
        # get the tag name
        'get-tagname',
        ('tagname',AllInCharSet,tagname_charset),
        # look for attributes
        'get-attributes',
        (None,AllInCharSet,white_charset,'incorrect-attributes'),
         (None,Is,'>',+1,MatchOk),
         ('tagattr',Table,tagattr),
         (None,Jump,To,-3),
         (None,Is,'>',+1,MatchOk),
        # handle incorrect attributes
        'incorrect-attributes',
        (error,AllNotIn,'> \n\r\t'),
        (None,Jump,To,'get-attributes')
        )

    print 'TagTable()'
    htmltable_tt = TagTable(htmltable)
    htmltag_tt = TagTable(htmltag)

    if HAVE_UNICODE:
        print 'UnicodeTagTable()'
        utt = UnicodeTagTable(htmltag)

    print 'TagTable() pickling'
    ptt = pickle.dumps(htmltable_tt)
    tt1 = pickle.loads(ptt)

    print 'TextSearch() pickling'
    pts = pickle.dumps(TextSearch('test'))
    ts1 = pickle.loads(pts)

    if 0:
        print 'HTML Table:'
        pprint.pprint(htmltable)
        print 'TagTable .dump() version of the HTML table:'
        pprint.pprint(htmltable_tt.dump())

    ###

    print 'TextSearch() object (Boyer-Moore)'
    ts = TextSearch('test')
    ts = TextSearch('test', None)
    ts = TextSearch('test', 'x'*256)
    ts = TextSearch('test', None, BOYERMOORE)
    
    ts = TextSearch('test')
    assert ts.search('    test') == (4, 8), ts.search('    test')
    assert ts.search('    test   ') == (4, 8)
    assert ts.search('    abc   ') == (0, 0)
    assert ts.find('    test') == 4, ts.find('    test')
    assert ts.find('    test   ') == 4
    assert ts.find('    abd   ') == -1
    assert ts.findall('    test  test  ') == [(4, 8), (10, 14)]
    assert ts.findall('   abc def  ') == []
    if HAVE_UNICODE:
        try:
            ts.search(unicode('    test'))
        except TypeError:
            pass
        else:
            raise AssertionError,'Boyer-Moore does not work with Unicode'
        try:
            ts.find(unicode('    test'))
        except TypeError:
            pass
        else:
            raise AssertionError,'Boyer-Moore does not work with Unicode'
        try:
            ts.findall(unicode('    test  test  '))
        except TypeError:
            pass
        else:
            raise AssertionError,'Boyer-Moore does not work with Unicode'

    try:
        ts = TextSearch('test', None, FASTSEARCH)
    except ValueError:
        pass
    else:
        print 'TextSearch() object (FastSearch)'
        assert ts.search('    test') == (4, 8)
        assert ts.search('    test   ') == (4, 8)
        assert ts.search('    abc   ') == (0, 0)
        assert ts.find('    test') == 4
        assert ts.find('    test   ') == 4
        assert ts.find('    abd   ') == -1
        assert ts.findall('    test  test  ') == [(4, 8), (10, 14)]
        assert ts.findall('   abc def  ') == []

    print 'TextSearch() object (Trivial)'
    ts = TextSearch('test', algorithm=TRIVIAL)
    assert ts.search('    test') == (4, 8)
    assert ts.search('    test   ') == (4, 8)
    assert ts.search('    abc   ') == (0, 0)
    assert ts.find('    test') == 4
    assert ts.find('    test   ') == 4
    assert ts.find('    abd   ') == -1
    assert ts.findall('    test  test  ') == [(4, 8), (10, 14)]
    assert ts.findall('   abc def  ') == []
    if HAVE_UNICODE:
        print 'TextSearch() object (Trivial; Unicode)'
        assert ts.search(unicode('    test')) == (4, 8)
        assert ts.search(unicode('    test   ')) == (4, 8)
        assert ts.search(unicode('    abc   ')) == (0, 0)
        assert ts.find(unicode('    test')) == 4
        assert ts.find(unicode('    test   ')) == 4
        assert ts.find(unicode('    abd   ')) == -1
        assert ts.findall(unicode('    test  test  ')) == [(4, 8), (10, 14)]
        assert ts.findall(unicode('   abc def  ')) == []
        ts = TextSearch(unicode('test'), algorithm=TRIVIAL)
        assert ts.search('    test') == (4, 8)
        assert ts.search('    test   ') == (4, 8)
        assert ts.search('    abc   ') == (0, 0)
        assert ts.find('    test') == 4
        assert ts.find('    test   ') == 4
        assert ts.find('    abd   ') == -1
        assert ts.findall('    test  test  ') == [(4, 8), (10, 14)]
        assert ts.findall('   abc def  ') == []
        assert ts.search(unicode('    test')) == (4, 8)
        assert ts.search(unicode('    test   ')) == (4, 8)
        assert ts.search(unicode('    abc   ')) == (0, 0)
        assert ts.find(unicode('    test')) == 4
        assert ts.find(unicode('    test   ')) == 4
        assert ts.find(unicode('    abd   ')) == -1
        assert ts.findall(unicode('    test  test  ')) == [(4, 8), (10, 14)]
        assert ts.findall(unicode('   abc def  ')) == []
        ts = TextSearch(unicode('test'))
        assert ts.algorithm == TRIVIAL

    print 'is_whitespace()'
    assert is_whitespace('   \t\r') == 1
    assert is_whitespace(' 123  ') == 0
    if HAVE_UNICODE:
        assert is_whitespace(unicode('   \t\r')) == 1
        assert is_whitespace(unicode(' 123  ')) == 0

    print 'collapse()'
    assert collapse('a\nb\nc') == 'a b c'
    assert collapse('a\nb\nc', '-') == 'a-b-c'
    if HAVE_UNICODE:
        assert collapse(unicode('a\nb\nä','latin-1')) == unicode('a b ä','latin-1')
        assert collapse(unicode('a\nb\nä','latin-1'), '-') == unicode('a-b-ä','latin-1')

    print 'splitwords()'
    assert splitwords('a b c') == ['a', 'b', 'c']
    if HAVE_UNICODE:
        assert splitwords(unicode('a b ä','latin-1')) == [ua, ub, unicode('ä','latin-1')]

    print 'countlines()'
    assert countlines('a\nb\nc') == 3
    assert countlines('a\nb\nc\n') == 3
    if HAVE_UNICODE:
        assert countlines(unicode('a\nb\nä','latin-1')) == 3

    print 'splitlines()'
    assert splitlines('a\nb\r\nc') == ['a', 'b', 'c']
    assert splitlines('a\nb\r\nc\r') == ['a', 'b', 'c']
    if HAVE_UNICODE:
        assert splitlines(unicode('a\nb\r\nä\r','latin-1')) == [ua, ub, unicode('ä','latin-1')]

    print 'replace()'
    assert replace('a\nb\nc', '\n', ' ') == 'a b c'
    assert replace('a\nb\nc', '\n', '-') == 'a-b-c'
    if HAVE_UNICODE:
        assert replace(unicode('a\nb\nä','latin-1'), '\n', ' ') == unicode('a b ä','latin-1')
        assert replace(unicode('a\nb\nä','latin-1'), '\n', '-') == unicode('a-b-ä','latin-1')

    print 'multireplace()'
    assert multireplace('a\nb\nc', [(' ', 1, 2)]) == 'a b\nc'
    assert multireplace('a\nb\nc', [('-', 1, 2), ('-', 3, 4)]) == 'a-b-c'
    if HAVE_UNICODE:
        assert multireplace(unicode('a\nb\nä','latin-1'), [(' ', 1, 2)]) == unicode('a b\nä','latin-1')
        assert multireplace(unicode('a\nb\nä','latin-1'), [('-', 1, 2), ('-', 3, 4)]) == unicode('a-b-ä','latin-1')

    print 'quoted_split()'
    assert quoted_split('  a, b  ,\t c,d ,e ,"ab,cd,de" ,\'a,b\'', ',') == \
           ['a', 'b', 'c', 'd', 'e', 'ab,cd,de', 'a,b']
    # twice to test table cache
    assert quoted_split('  a, b  ,\t c,d ,e ,"ab,cd,de" ,\'a,b\'', ',') == \
           ['a', 'b', 'c', 'd', 'e', 'ab,cd,de', 'a,b']
    assert quoted_split('  a b  \t c d e "ab cd de" \'a b\'') == \
           ['a', 'b', 'c', 'd', 'e', 'ab cd de', 'a b']
    assert quoted_split(',,a', ',') == ['', '', 'a']
    assert quoted_split(',,a,', ',') == ['', '', 'a', '']
    if HAVE_UNICODE:
        assert quoted_split(unicode('  a, b  ,\t c,d ,e ,"ab,cd,de" ,\'a,b\''), ',') == \
               [ua, ub, uc, ud, ue, unicode('ab,cd,de'), unicode('a,b')]
        assert quoted_split(unicode(',,a'), ',') == [uempty, uempty, ua]
        assert quoted_split(unicode(',,a,'), ',') == [uempty, uempty, ua, uempty]

    # Clear the TagTable cache
    tagtable_cache.clear()

    break

print
print 'Works.'



""" mxProxy - Proxy wrapper

    Copyright (c) 2000, Marc-Andre Lemburg; mailto:mal@lemburg.com
    Copyright (c) 2000-2013, eGenix.com Software GmbH; mailto:info@egenix.com
    See the documentation for further information on copyrights,
    or contact the author. All Rights Reserved.

"""
from mxProxy import *
from mxProxy import __version__



#!/usr/local/bin/python

""" RegExp - tag a string using regexps (Version 0.1)
    
    Copyright (c) 2000, Marc-Andre Lemburg; mailto:mal@lemburg.com
    Copyright (c) 2000-2013, eGenix.com Software GmbH; mailto:info@egenix.com
    See the documentation for further information on copyrights,
    or contact the author. All Rights Reserved.
"""

# engine + constants
from mx.TextTools import *

# special matching function
def match_regexp(text,position,len_text,regexpr):
    position = position + max(0,regexpr.match(text,position).end())
    return position

# create a table in the sense re_1\|re_2\|...\|re_N where
# re_i \in regexps
def or_regexps(regexps):
    # regexps = list of compiled regexps
    l = []
    for i in range(len(regexps)):
        l.append((i,CallArg,(match_regexp,regexps[i]),+1,MatchOk))
    l.append((None,Fail,Here))
    return tuple(l)

if __name__ == '__main__':

    # create some simple regexps
    import re
    regexps = [ 'spam*', 'ham*', 'eggs' ]
    regexps = map(re.compile, regexps)
    table = or_regexps(regexps)

    text = raw_input('type some words: ')
    result, taglist, nextindex = tag(text,table)

    if result:
        print 'subexpr nr.',taglist[0][0],'matched:',taglist[0]
    else:
        print 'no match'

    if nextindex < len(text): 
        print 'rest of unparsed input:',text[nextindex:]



""" NewBuiltins - Auto-install the collection of new builtins provided
                  by mx.Tools

    Redirection module; use mx.Tools.NewBuiltins instead !!!

    Copyright (c) 2000, Marc-Andre Lemburg; mailto:mal@lemburg.com
    Copyright (c) 2000-2013, eGenix.com Software GmbH; mailto:info@egenix.com
    See the documentation for further information on copyrights,
    or contact the author. All Rights Reserved.
"""
import mx.Tools.NewBuiltins
